# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.modules sage.rings.finite_rings
r"""
Gabidulin Code

This module provides the :class:`~sage.coding.gabidulin.GabidulinCode`, which constructs
Gabidulin Codes that are the rank metric equivalent of Reed Solomon codes and are
defined as the evaluation codes of degree-restricted skew polynomials.

This module also provides :class:`~sage.coding.gabidulin.GabidulinPolynomialEvaluationEncoder`,
an encoder with a skew polynomial message space and :class:`~sage.coding.gabidulin.GabidulinVectorEvaluationEncoder`,
an encoder based on the generator matrix. It also provides a decoder
:class:`~sage.coding.gabidulin.GabidulinGaoDecoder` which corrects errors using
the Gao algorithm in the rank metric.

AUTHOR:

- Arpit Merchant (2016-08-16)
- Marketa Slukova (2019-08-19): initial version
"""
from sage.matrix.constructor import matrix
from sage.modules.free_module_element import vector
from sage.coding.encoder import Encoder
from sage.coding.decoder import Decoder, DecodingError
from sage.coding.linear_rank_metric import AbstractLinearRankMetricCode
from sage.categories.fields import Fields


class GabidulinCode(AbstractLinearRankMetricCode):
    r"""
    A Gabidulin Code.

    DEFINITION:

    A linear Gabidulin code Gab[n, k] over `F_{q^m}` of length `n` (at most
    `m`) and dimension `k` (at most `n`) is the set of all codewords, that
    are the evaluation of a `q`-degree restricted skew polynomial `f(x)`
    belonging to the skew polynomial constructed over the base ring `F_{q^m}`
    and the twisting homomorphism `\sigma`.

    .. math::

        \{ \text{Gab[n, k]} = \big\{ (f(g_0) f(g_1) ... f(g_{n-1})) = f(\textbf{g}) : \text{deg}_{q}f(x) < k \big\} \}

    where the fixed evaluation points `g_0, g_1,..., g_{n-1}` are linearly
    independent over `F_{q^m}`.

    EXAMPLES:

    A Gabidulin Code can be constructed in the following way::

        sage: Fqm = GF(16)
        sage: Fq = GF(4)
        sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq)
        sage: C
        [2, 2, 1] linear Gabidulin code over GF(16)/GF(4)
    """
    _registered_encoders = {}
    _registered_decoders = {}

    def __init__(self, base_field, length, dimension, sub_field=None,
            twisting_homomorphism=None, evaluation_points=None):
        r"""
        Representation of a Gabidulin Code.

        INPUT:

        - ``base_field`` -- finite field of order `q^m` where `q` is a prime power
          and `m` is an integer

        - ``length`` -- length of the resulting code

        - ``dimension`` -- dimension of the resulting code

        - ``sub_field`` -- (default: ``None``) finite field of order `q`
          which is a subfield of the ``base_field``. If not given, it is the
          prime subfield of the ``base_field``.

        - ``twisting_homomorphism`` -- (default: ``None``) homomorphism of the
          underlying skew polynomial ring. If not given, it is the Frobenius
          endomorphism on ``base_field``, which sends an element `x` to `x^{q}`.

        - ``evaluation_points`` -- (default: ``None``) list of elements
          `g_0, g_1,...,g_{n-1}` of the ``base_field`` that are linearly
          independent over the ``sub_field``. These elements form the first row
          of the generator matrix. If not specified, these are the `nth` powers
          of the generator of the ``base_field``.

        Both parameters ``sub_field`` and ``twisting_homomorphism`` are optional.
        Since they are closely related, here is a complete list of behaviours:

        - both ``sub_field`` and ``twisting_homomorphism`` given -- in this case
          we only check that given that ``twisting_homomorphism`` has a fixed
          field method, it returns ``sub_field``

        - only ``twisting_homomorphism`` given -- we set ``sub_field`` to be the
          fixed field of the ``twisting_homomorphism``. If such method does not
          exist, an error is raised.

        - only ``sub_field`` given -- we set ``twisting_homomorphism`` to be the
          Frobenius of the field extension

        - neither ``sub_field`` or ``twisting_homomorphism`` given -- we take
          ``sub_field`` to be the prime field of ``base_field`` and the
          ``twisting_homomorphism`` to be the Frobenius wrt. the prime field

        TESTS:

        If ``length`` is bigger than the degree of the extension, an error is
        raised::

            sage: C = codes.GabidulinCode(GF(64), 4, 3, GF(4))
            Traceback (most recent call last):
            ...
            ValueError: 'length' can be at most the degree of the extension, 3

        If the number of evaluation points is not equal to the length
        of the code, an error is raised::

            sage: Fqm = GF(5^20)
            sage: Fq = GF(5)
            sage: aa = Fqm.gen()
            sage: evals = [ aa^i for i in range(21) ]
            sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq, None, evals)
            Traceback (most recent call last):
            ...
            ValueError: the number of evaluation points should be equal to the length of the code

        If evaluation points are not linearly independent over the ``base_field``,
        an error is raised::

            sage: evals = [ aa*i for i in range(2) ]
            sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq, None, evals)
            Traceback (most recent call last):
            ...
            ValueError: the evaluation points provided are not linearly independent

        If an evaluation point does not belong to the ``base_field``, an error
        is raised::

            sage: a = GF(3).gen()
            sage: evals = [ a*i for i in range(2) ]
            sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq, None, evals)
            Traceback (most recent call last):
            ...
            ValueError: evaluation point does not belong to the 'base field'

        Given that both ``sub_field`` and ``twisting_homomorphism`` are specified
        and ``twisting_homomorphism`` has a fixed field method. If the fixed
        field of ``twisting_homomorphism`` is not ``sub_field``, an error is
        raised::

            sage: Fqm = GF(64)
            sage: Fq = GF(8)
            sage: twist = GF(64).frobenius_endomorphism(n=2)
            sage: C = codes.GabidulinCode(Fqm, 3, 2, Fq, twist)
            Traceback (most recent call last):
            ...
            ValueError: the fixed field of the twisting homomorphism has to be the relative field of the extension

        If ``twisting_homomorphism`` is given, but ``sub_field`` is not. In case
        ``twisting_homomorphism`` does not have a fixed field method, and error
        is raised::

            sage: Fqm.<z6> = GF(64)
            sage: sigma = Hom(Fqm, Fqm)[1]; sigma
            Ring endomorphism of Finite Field in z6 of size 2^6
                Defn: z6 |--> z6^2
            sage: C = codes.GabidulinCode(Fqm, 3, 2, None, sigma)
            Traceback (most recent call last):
            ...
            ValueError: if 'sub_field' is not given, the twisting homomorphism has to have a 'fixed_field' method
        """
        twist_fix_field = None
        have_twist = (twisting_homomorphism is not None)
        have_subfield = (sub_field is not None)

        if have_twist and have_subfield:
            try:
                twist_fix_field = twisting_homomorphism.fixed_field()[0]
            except AttributeError:
                pass
            if twist_fix_field and twist_fix_field.order() != sub_field.order():
                raise ValueError("the fixed field of the twisting homomorphism has to be the relative field of the extension")

        if have_twist and not have_subfield:
            if not twist_fix_field:
                raise ValueError("if 'sub_field' is not given, the twisting homomorphism has to have a 'fixed_field' method")
            else:
                sub_field = twist_fix_field

        if (not have_twist) and have_subfield:
            twisting_homomorphism = base_field.frobenius_endomorphism(n=sub_field.degree())

        if (not have_twist) and not have_subfield:
            sub_field = base_field.base_ring()
            twisting_homomorphism = base_field.frobenius_endomorphism()

        self._twisting_homomorphism = twisting_homomorphism

        super().__init__(base_field, sub_field, length, "VectorEvaluation", "Gao")

        if length > self.extension_degree():
            raise ValueError("'length' can be at most the degree of the extension, {}".format(self.extension_degree()))
        if evaluation_points is None:
            evaluation_points = [base_field.gen()**i for i in range(base_field.degree())][:length]
        else:
            if not len(evaluation_points) == length:
                raise ValueError("the number of evaluation points should be equal to the length of the code")
            for i in range(length):
                if evaluation_points[i] not in base_field:
                    raise ValueError("evaluation point does not belong to the 'base field'")
            basis = self.matrix_form_of_vector(vector(evaluation_points))
            if basis.rank() != length:
                raise ValueError("the evaluation points provided are not linearly independent")
        self._evaluation_points = evaluation_points
        self._dimension = dimension

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: Fqm = GF(16)
            sage: Fq = GF(4)
            sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq); C
            [2, 2, 1] linear Gabidulin code over GF(16)/GF(4)
        """
        R = self.base_field()
        S = self.sub_field()
        if R and S in Fields():
            return "[%s, %s, %s] linear Gabidulin code over GF(%s)/GF(%s)" % (self.length(), self.dimension(), self.minimum_distance(), R.cardinality(), S.cardinality())
        else:
            return "[%s, %s, %s] linear Gabidulin code over %s/%s" % (self.length(), self.dimension(), self.minimum_distance(), R, S)

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: Fqm = GF(16)
            sage: Fq = GF(4)
            sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq);
            sage: latex(C)
            [2, 2, 1] \textnormal{ linear Gabidulin code over } \Bold{F}_{2^{4}}/\Bold{F}_{2^{2}}
        """
        txt = "[%s, %s, %s] \\textnormal{ linear Gabidulin code over } %s/%s"
        return txt % (self.length(), self.dimension(), self.minimum_distance(),
                      self.base_field()._latex_(), self.sub_field()._latex_())

    def __eq__(self, other):
        """
        Test equality between Gabidulin Code objects.

        INPUT:

        - ``other`` -- another Gabidulin Code object

        OUTPUT: boolean

        EXAMPLES::

            sage: Fqm = GF(16)
            sage: Fq = GF(4)
            sage: C1 = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: C2 = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: C1.__eq__(C2)
            True

            sage: Fqmm = GF(64)
            sage: C3 = codes.GabidulinCode(Fqmm, 2, 2, Fq)
            sage: C3.__eq__(C2)
            False
        """
        return isinstance(other, GabidulinCode) \
            and self.base_field() == other.base_field() \
            and self.sub_field() == other.sub_field() \
            and self.length() == other.length() \
            and self.dimension() == other.dimension() \
            and self.evaluation_points() == other.evaluation_points()

    def twisting_homomorphism(self):
        r"""
        Return the twisting homomorphism of ``self``.

        EXAMPLES::

            sage: Fqm = GF(5^20)
            sage: Fq = GF(5^4)
            sage: C = codes.GabidulinCode(Fqm, 5, 3, Fq)
            sage: C.twisting_homomorphism()
            Frobenius endomorphism z20 |--> z20^(5^4) on Finite Field in z20 of size 5^20
        """
        return self._twisting_homomorphism

    def minimum_distance(self):
        r"""
        Return the minimum distance of ``self``.

        Since Gabidulin Codes are Maximum-Distance-Separable (MDS), this returns
         ``self.length() - self.dimension() + 1``.

        EXAMPLES::

            sage: Fqm = GF(5^20)
            sage: Fq = GF(5)
            sage: C = codes.GabidulinCode(Fqm, 20, 15, Fq)
            sage: C.minimum_distance()
            6
        """
        return self.length() - self.dimension() + 1

    def parity_evaluation_points(self):
        r"""
        Return the parity evaluation points of ``self``.

        These form the first row of the parity check matrix of ``self``.

        EXAMPLES::

            sage: C = codes.GabidulinCode(GF(2^10), 5, 2)
            sage: list(C.parity_check_matrix().row(0)) == C.parity_evaluation_points() #indirect_doctest
            True
        """
        eval_pts = self.evaluation_points()
        n = self.length()
        k = self.dimension()
        sigma = self.twisting_homomorphism()

        coefficient_matrix = matrix(self.base_field(), n - 1, n,
                                    lambda i, j: (sigma**(-n + k + 1 + i))(eval_pts[j]))
        solution_space = coefficient_matrix.right_kernel()
        return list(solution_space.basis()[0])

    def dual_code(self):
        r"""
        Return the dual code `C^{\perp}` of ``self``, the code `C`,

        .. MATH::

            C^{\perp} = \{ v \in V\ |\ v\cdot c = 0,\ \forall c \in C \}.

        EXAMPLES::

            sage: C = codes.GabidulinCode(GF(2^10), 5, 2)
            sage: C1 = C.dual_code(); C1
            [5, 3, 3] linear Gabidulin code over GF(1024)/GF(2)
            sage: C == C1.dual_code()
            True
        """
        return GabidulinCode(self.base_field(), self.length(),
                             self.length() - self.dimension(),
                             self.sub_field(),
                             self.twisting_homomorphism(),
                             self.parity_evaluation_points())

    def parity_check_matrix(self):
        r"""
        Return the parity check matrix of ``self``.

        This is the generator matrix of the dual code of ``self``.

        EXAMPLES::

            sage: C = codes.GabidulinCode(GF(2^3), 3, 2)
            sage: C.parity_check_matrix()
            [        1        z3 z3^2 + z3]
            sage: C.parity_check_matrix() == C.dual_code().generator_matrix()
            True
        """
        return self.dual_code().generator_matrix()

    def evaluation_points(self):
        """
        Return the evaluation points of ``self``.

        EXAMPLES::

            sage: Fqm = GF(5^20)
            sage: Fq = GF(5^4)
            sage: C = codes.GabidulinCode(Fqm, 4, 4, Fq)
            sage: C.evaluation_points()
            [1, z20, z20^2, z20^3]
        """
        return self._evaluation_points

# ---------------------- encoders ------------------------------


class GabidulinVectorEvaluationEncoder(Encoder):

    def __init__(self, code):
        """
        This method constructs the vector evaluation encoder for
        Gabidulin Codes.

        INPUT:

        - ``code`` -- the associated code of this encoder

        EXAMPLES::

            sage: Fqm = GF(16)
            sage: Fq = GF(4)
            sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: E = codes.encoders.GabidulinVectorEvaluationEncoder(C)
            sage: E
            Vector evaluation style encoder for [2, 2, 1] linear Gabidulin code over GF(16)/GF(4)

        Alternatively, we can construct the encoder from ``C`` directly::

            sage: E = C.encoder("VectorEvaluation")
            sage: E
            Vector evaluation style encoder for [2, 2, 1] linear Gabidulin code over GF(16)/GF(4)

        TESTS:

        If the code is not a Gabidulin code, an error is raised::

            sage: C = codes.HammingCode(GF(4), 2)
            sage: E = codes.encoders.GabidulinVectorEvaluationEncoder(C)
            Traceback (most recent call last):
            ...
            ValueError: code has to be a Gabidulin code
        """
        if not isinstance(code, GabidulinCode):
            raise ValueError("code has to be a Gabidulin code")
        super().__init__(code)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: Fqm = GF(5^20)
            sage: Fq = GF(5^4)
            sage: C = codes.GabidulinCode(Fqm, 4, 4, Fq)
            sage: E = codes.encoders.GabidulinVectorEvaluationEncoder(C); E
            Vector evaluation style encoder for [4, 4, 1] linear Gabidulin code over GF(95367431640625)/GF(625)
        """
        return "Vector evaluation style encoder for %s" % self.code()

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: Fqm = GF(5^20)
            sage: Fq = GF(5^4)
            sage: C = codes.GabidulinCode(Fqm, 4, 4, Fq)
            sage: E = codes.encoders.GabidulinVectorEvaluationEncoder(C)
            sage: latex(E)
            \textnormal{Vector evaluation style encoder for } [4, 4, 1] \textnormal{ linear Gabidulin code over } \Bold{F}_{5^{20}}/\Bold{F}_{5^{4}}
        """
        return "\\textnormal{Vector evaluation style encoder for } %s" % self.code()._latex_()

    def __eq__(self, other):
        """
        Test equality between Gabidulin Generator Matrix Encoder objects.

        INPUT:

        - ``other`` -- another Gabidulin Generator Matrix Encoder

        OUTPUT: boolean

        EXAMPLES::

            sage: Fqm = GF(16)
            sage: Fq = GF(4)
            sage: C1 = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: E1 = codes.encoders.GabidulinVectorEvaluationEncoder(C1)
            sage: C2 = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: E2 = codes.encoders.GabidulinVectorEvaluationEncoder(C2)
            sage: E1.__eq__(E2)
            True

            sage: Fqmm = GF(64)
            sage: C3 = codes.GabidulinCode(Fqmm, 2, 2, Fq)
            sage: E3 = codes.encoders.GabidulinVectorEvaluationEncoder(C3)
            sage: E3.__eq__(E2)
            False
        """
        return isinstance(other, GabidulinVectorEvaluationEncoder) \
            and self.code() == other.code()

    def generator_matrix(self):
        """
        Return the generator matrix of ``self``.

        EXAMPLES::

            sage: Fqm = GF(2^9)
            sage: Fq = GF(2^3)
            sage: C = codes.GabidulinCode(Fqm, 3, 3, Fq)
            sage: (list(C.generator_matrix().row(1))
            ....:   == [C.evaluation_points()[i]**(2**3) for i in range(3)])
            True
        """
        from functools import reduce
        C = self.code()
        eval_pts = C.evaluation_points()
        sigma = C.twisting_homomorphism()

        def create_matrix_elements(A, k, f):
            return reduce(lambda L, x: [x] +
                          [list(map(f, l)) for l in L], [A] * k, [])
        return matrix(C.base_field(), C.dimension(), C.length(),
                      create_matrix_elements(eval_pts, C.dimension(), sigma))


class GabidulinPolynomialEvaluationEncoder(Encoder):
    r"""
    Encoder for Gabidulin codes which uses evaluation of skew polynomials to
    obtain codewords.

    Let `C` be a Gabidulin code of length `n` and dimension `k` over some
    finite field `F = GF(q^m)`. We denote by `\alpha_i` its evaluations
    points, where `1 \leq i \leq n`. Let `p`, a skew polynomial of degree at
    most `k-1` in `F[x]`, be the message.

    The encoding of `m` will be the following codeword:

    .. MATH::

        (p(\alpha_1), \dots, p(\alpha_n)).

    TESTS:

    This module uses the following experimental feature.
    This test block is here only to trigger the experimental warning so it does not
    interferes with doctests::

        sage: Fqm = GF(2^9)
        sage: Fq = GF(2^3)
        sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq)
        sage: S.<x> = Fqm['x', C.twisting_homomorphism()]
        sage: z9 = Fqm.gen()
        sage: p = (z9^6 + z9^2 + z9 + 1)*x + z9^7 + z9^5 + z9^4 + z9^2
        sage: vector(p.multi_point_evaluation(C.evaluation_points()))
        doctest:...: FutureWarning: This class/method/function is marked as experimental.
        It, its functionality or its interface might change without a formal deprecation.
        See https://github.com/sagemath/sage/issues/13215 for details.
        (z9^7 + z9^6 + z9^5 + z9^4 + z9 + 1, z9^6 + z9^5 + z9^3 + z9)

    EXAMPLES::

        sage: Fqm = GF(16)
        sage: Fq = GF(4)
        sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq)
        sage: E = codes.encoders.GabidulinPolynomialEvaluationEncoder(C)
        sage: E
        Polynomial evaluation style encoder for
         [2, 2, 1] linear Gabidulin code over GF(16)/GF(4)

    Alternatively, we can construct the encoder from ``C`` directly::

        sage: E = C.encoder("PolynomialEvaluation")
        sage: E
        Polynomial evaluation style encoder for
         [2, 2, 1] linear Gabidulin code over GF(16)/GF(4)
    """

    def __init__(self, code):
        r"""
        INPUT:

        - ``code`` -- the associated code of this encoder

        TESTS:

        If the code is not a Gabidulin code, an error is raised::

            sage: C = codes.HammingCode(GF(4), 2)
            sage: E = codes.encoders.GabidulinPolynomialEvaluationEncoder(C)
            Traceback (most recent call last):
            ...
            ValueError: code has to be a Gabidulin code
        """
        if not isinstance(code, GabidulinCode):
            raise ValueError("code has to be a Gabidulin code")
        super().__init__(code)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: Fqm = GF(5^20)
            sage: Fq = GF(5^4)
            sage: C = codes.GabidulinCode(Fqm, 4, 4, Fq)
            sage: E = codes.encoders.GabidulinPolynomialEvaluationEncoder(C); E
            Polynomial evaluation style encoder for [4, 4, 1] linear Gabidulin code over GF(95367431640625)/GF(625)
        """
        return "Polynomial evaluation style encoder for %s" % self.code()

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: Fqm = GF(5^20)
            sage: Fq = GF(5^4)
            sage: C = codes.GabidulinCode(Fqm, 4, 4, Fq)
            sage: E = codes.encoders.GabidulinPolynomialEvaluationEncoder(C)
            sage: latex(E)
            \textnormal{Polynomial evaluation style encoder for } [4, 4, 1] \textnormal{ linear Gabidulin code over } \Bold{F}_{5^{20}}/\Bold{F}_{5^{4}}
        """
        return "\\textnormal{Polynomial evaluation style encoder for } %s" % self.code()._latex_()

    def __eq__(self, other):
        """
        Test equality between Gabidulin Polynomial Evaluation Encoder objects.

        INPUT:

        - ``other`` -- another Gabidulin Polynomial Evaluation Encoder

        OUTPUT: boolean

        EXAMPLES::

            sage: Fqm = GF(16)
            sage: Fq = GF(4)
            sage: C1 = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: E1 = codes.encoders.GabidulinPolynomialEvaluationEncoder(C1)
            sage: C2 = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: E2 = codes.encoders.GabidulinPolynomialEvaluationEncoder(C2)
            sage: E1.__eq__(E2)
            True

            sage: Fqmm = GF(64)
            sage: C3 = codes.GabidulinCode(Fqmm, 2, 2, Fq)
            sage: E3 = codes.encoders.GabidulinPolynomialEvaluationEncoder(C3)
            sage: E3.__eq__(E2)
            False
        """
        return isinstance(other, GabidulinPolynomialEvaluationEncoder) \
            and self.code() == other.code()

    def message_space(self):
        r"""
        Return the message space of the associated code of ``self``.

        EXAMPLES::

            sage: Fqm = GF(5^20)
            sage: Fq = GF(5^4)
            sage: C = codes.GabidulinCode(Fqm, 4, 4, Fq)
            sage: E = codes.encoders.GabidulinPolynomialEvaluationEncoder(C)
            sage: E.message_space()
            Ore Polynomial Ring in x over Finite Field in z20 of size 5^20
             twisted by z20 |--> z20^(5^4)
        """
        C = self.code()
        return C.base_field()['x', C.twisting_homomorphism()]

    def encode(self, p, form='vector'):
        """
        Transform the polynomial ``p`` into a codeword of :meth:`code`.

        The output codeword can be represented as a vector or a matrix,
        depending on the ``form`` input.

        INPUT:

        - ``p`` -- a skew polynomial from the message space of ``self`` of degree
          less than ``self.code().dimension()``

        - ``form`` -- type parameter taking strings "vector" or "matrix"
          as values and converting the output codeword into the respective form
          (default: ``'vector'``)

        OUTPUT: a codeword corresponding to `p` in vector or matrix form

        EXAMPLES::

            sage: Fqm = GF(2^9)
            sage: Fq = GF(2^3)
            sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: E = codes.encoders.GabidulinPolynomialEvaluationEncoder(C)
            sage: S.<x> = Fqm['x', C.twisting_homomorphism()]
            sage: z9 = Fqm.gen()
            sage: p = (z9^6 + z9^2 + z9 + 1)*x + z9^7 + z9^5 + z9^4 + z9^2
            sage: codeword_vector = E.encode(p, "vector"); codeword_vector
            (z9^7 + z9^6 + z9^5 + z9^4 + z9 + 1, z9^6 + z9^5 + z9^3 + z9)
            sage: codeword_matrix = E.encode(p, "matrix"); codeword_matrix
            [           z3     z3^2 + z3]
            [           z3             1]
            [         z3^2 z3^2 + z3 + 1]

        TESTS:

        If the skew polynomial, `p`, has degree greater than or equal to the
        dimension of the code, an error is raised::

            sage: t = z9^4*x^2 + z9
            sage: codeword_vector = E.encode(t, "vector"); codeword_vector
            Traceback (most recent call last):
            ...
            ValueError: the skew polynomial to encode must have degree at most 1

        The skew polynomial, `p`, must belong to the message space of the code.
        Otherwise, an error is raised::

            sage: Fqmm = GF(2^12)
            sage: S.<x> = Fqmm['x', Fqmm.frobenius_endomorphism(n=3)]
            sage: q = S.random_element(degree=2)
            sage: codeword_vector = E.encode(q, "vector"); codeword_vector
            Traceback (most recent call last):
            ...
            ValueError: the message to encode must be in Ore Polynomial Ring in x over Finite Field in z9 of size 2^9 twisted by z9 |--> z9^(2^3)
        """
        C = self.code()
        M = self.message_space()
        if p not in M:
            raise ValueError("the message to encode must be in %s" % M)
        if p.degree() >= C.dimension():
            raise ValueError("the skew polynomial to encode must have degree at most %s" % (C.dimension() - 1))
        eval_pts = C.evaluation_points()
        codeword = p.multi_point_evaluation(eval_pts)
        if form == "vector":
            return vector(codeword)
        elif form == "matrix":
            return C.matrix_form_of_vector(vector(codeword))
        else:
            return ValueError("the argument 'form' takes only either 'vector' or 'matrix' as valid input")

    def unencode_nocheck(self, c):
        """
        Return the message corresponding to the codeword ``c``.

        Use this method with caution: it does not check if ``c``
        belongs to the code, and if this is not the case, the output is
        unspecified.

        INPUT:

        - ``c`` -- a codeword of :meth:`code`

        OUTPUT:

        - a skew polynomial of degree less than ``self.code().dimension()``

        EXAMPLES::

            sage: Fqm = GF(2^9)
            sage: Fq = GF(2^3)
            sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: E = codes.encoders.GabidulinPolynomialEvaluationEncoder(C)
            sage: S.<x> = Fqm['x', C.twisting_homomorphism()]
            sage: z9 = Fqm.gen()
            sage: p = (z9^6 + z9^4)*x + z9^2 + z9
            sage: codeword_vector = E.encode(p, "vector")
            sage: E.unencode_nocheck(codeword_vector)
            (z9^6 + z9^4)*x + z9^2 + z9
        """
        C = self.code()
        eval_pts = C.evaluation_points()
        values = [c[i] for i in range(len(c))]
        points = [(eval_pts[i], values[i]) for i in range(len(eval_pts))]
        p = self.message_space().lagrange_polynomial(points)
        return p


# ---------------------- decoders ------------------------------


class GabidulinGaoDecoder(Decoder):

    def __init__(self, code):
        r"""
        Gao style decoder for Gabidulin Codes.

        INPUT:

        - ``code`` -- the associated code of this decoder

        EXAMPLES::

            sage: Fqm = GF(16)
            sage: Fq = GF(4)
            sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: D = codes.decoders.GabidulinGaoDecoder(C)
            sage: D
            Gao decoder for [2, 2, 1] linear Gabidulin code over GF(16)/GF(4)

        Alternatively, we can construct the encoder from ``C`` directly::

            sage: D = C.decoder("Gao")
            sage: D
            Gao decoder for [2, 2, 1] linear Gabidulin code over GF(16)/GF(4)

        TESTS:

        If the code is not a Gabidulin code, an error is raised::

            sage: C = codes.HammingCode(GF(4), 2)
            sage: D = codes.decoders.GabidulinGaoDecoder(C)
            Traceback (most recent call last):
            ...
            ValueError: code has to be a Gabidulin code
        """
        if not isinstance(code, GabidulinCode):
            raise ValueError("code has to be a Gabidulin code")
        super().__init__(code, code.ambient_space(), "PolynomialEvaluation")

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: Fqm = GF(5^20)
            sage: Fq = GF(5^4)
            sage: C = codes.GabidulinCode(Fqm, 4, 4, Fq)
            sage: D = codes.decoders.GabidulinGaoDecoder(C); D
            Gao decoder for [4, 4, 1] linear Gabidulin code over GF(95367431640625)/GF(625)
        """
        return "Gao decoder for %s" % self.code()

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: Fqm = GF(5^20)
            sage: Fq = GF(5^4)
            sage: C = codes.GabidulinCode(Fqm, 4, 4, Fq)
            sage: D = codes.decoders.GabidulinGaoDecoder(C)
            sage: latex(D)
            \textnormal{Gao decoder for } [4, 4, 1] \textnormal{ linear Gabidulin code over } \Bold{F}_{5^{20}}/\Bold{F}_{5^{4}}
        """
        return "\\textnormal{Gao decoder for } %s" % self.code()._latex_()

    def __eq__(self, other) -> bool:
        """
        Test equality between Gabidulin Gao Decoder objects.

        INPUT:

        - ``other`` -- another Gabidulin Gao Decoder

        OUTPUT: boolean

        EXAMPLES::

            sage: Fqm = GF(16)
            sage: Fq = GF(4)
            sage: C1 = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: D1 = codes.decoders.GabidulinGaoDecoder(C1)
            sage: C2 = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: D2 = codes.decoders.GabidulinGaoDecoder(C2)
            sage: D1.__eq__(D2)
            True

            sage: Fqmm = GF(64)
            sage: C3 = codes.GabidulinCode(Fqmm, 2, 2, Fq)
            sage: D3 = codes.decoders.GabidulinGaoDecoder(C3)
            sage: D3.__eq__(D2)
            False
        """
        return isinstance(other, GabidulinGaoDecoder) \
            and self.code() == other.code()

    def _partial_xgcd(self, a, b, d_stop):
        """
        Compute the partial gcd of `a` and `b` using the right linearized
        extended Euclidean algorithm up to the `d_stop` iterations. This
        is a private method for internal use only.

        INPUT:

        - ``a`` -- a skew polynomial

        - ``b`` -- another skew polynomial

        - ``d_stop`` -- the number of iterations for which the algorithm
          is to be run

        OUTPUT: ``r_c`` -- right linearized remainder of `a` and `b`

        - ``u_c`` -- right linearized quotient of `a` and `b`

        EXAMPLES::

            sage: Fqm = GF(2^9)
            sage: Fq = GF(2^3)
            sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: D = codes.decoders.GabidulinGaoDecoder(C)
            sage: E = codes.encoders.GabidulinPolynomialEvaluationEncoder(C)
            sage: S.<x> = Fqm['x', C.twisting_homomorphism()]
            sage: z9 = Fqm.gen()
            sage: p = (z9^6 + z9^4)*x + z9^2 + z9
            sage: codeword_vector = E.encode(p, "vector")
            sage: r = D.decode_to_message(codeword_vector) #indirect_doctest
            sage: r
            (z9^6 + z9^4)*x + z9^2 + z9
        """
        S = self.message_space()
        if (a not in S) or (b not in S):
            raise ValueError("both the input polynomials must belong to %s" % S)
        if a.degree() < b.degree():
            raise ValueError("degree of first polynomial must be greater than or equal to degree of second polynomial")
        r_p = a
        r_c = b
        u_p = S.zero()
        u_c = S.one()
        v_p = u_c
        v_c = u_p

        while r_c.degree() >= d_stop:
            (q, r_c), r_p = r_p.right_quo_rem(r_c), r_c
            u_c, u_p = u_p - q * u_c, u_c
            v_c, v_p = v_p - q * v_c, v_c
        return r_c, u_c

    def _decode_to_code_and_message(self, r):
        """
        Return the decoded codeword and message (skew polynomial)
        corresponding to the received codeword `r`. This is a
        private method for internal use only.

        INPUT:

        - ``r`` -- received codeword

        OUTPUT:

        - the decoded codeword and decoded message corresponding to
          the received codeword `r`

        EXAMPLES::

            sage: Fqm = GF(2^9)
            sage: Fq = GF(2^3)
            sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: D = codes.decoders.GabidulinGaoDecoder(C)
            sage: E = codes.encoders.GabidulinPolynomialEvaluationEncoder(C)
            sage: S.<x> = Fqm['x', C.twisting_homomorphism()]
            sage: z9 = Fqm.gen()
            sage: p = (z9^6 + z9^4)*x + z9^2 + z9
            sage: codeword_vector = E.encode(p, "vector")
            sage: r = D.decode_to_message(codeword_vector) #indirect doctest
            sage: r
            (z9^6 + z9^4)*x + z9^2 + z9
        """
        C = self.code()
        length = len(r)
        eval_pts = C.evaluation_points()
        S = self.message_space()

        if length == C.dimension() or r in C:
            return r, self.connected_encoder().unencode_nocheck(r)

        points = [(eval_pts[i], r[i]) for i in range(len(eval_pts))]
        # R = S.lagrange_polynomial(eval_pts, list(r))
        R = S.lagrange_polynomial(points)
        r_out, u_out = self._partial_xgcd(S.minimal_vanishing_polynomial(eval_pts),
                R, (C.length() + C.dimension()) // 2)
        quo, rem = r_out.left_quo_rem(u_out)
        if not rem.is_zero():
            raise DecodingError("Decoding failed because the number of errors exceeded the decoding radius")
        if quo not in S:
            raise DecodingError("Decoding failed because the number of errors exceeded the decoding radius")
        c = self.connected_encoder().encode(quo)
        if C.rank_weight_of_vector(c - r) > self.decoding_radius():
            raise DecodingError("Decoding failed because the number of errors exceeded the decoding radius")
        return c, quo

    def decode_to_code(self, r):
        """
        Return the decoded codeword corresponding to the
        received word `r`.

        INPUT:

        - ``r`` -- received codeword

        OUTPUT: the decoded codeword corresponding to the received codeword

        EXAMPLES::

            sage: Fqm = GF(3^20)
            sage: Fq = GF(3)
            sage: C = codes.GabidulinCode(Fqm, 5, 3, Fq)
            sage: D = codes.decoders.GabidulinGaoDecoder(C)
            sage: E = codes.encoders.GabidulinPolynomialEvaluationEncoder(C)
            sage: S.<x> = Fqm['x', C.twisting_homomorphism()]
            sage: z20 = Fqm.gen()
            sage: p = x
            sage: codeword_vector = E.encode(p, "vector")
            sage: codeword_vector
            (1, z20^3, z20^6, z20^9, z20^12)
            sage: l = list(codeword_vector)
            sage: l[0] = l[1] #make an error
            sage: D.decode_to_code(vector(l))
            (1, z20^3, z20^6, z20^9, z20^12)
        """
        return self._decode_to_code_and_message(r)[0]

    def decode_to_message(self, r):
        """
        Return the skew polynomial (message) corresponding to the
        received word `r`.

        INPUT:

        - ``r`` -- received codeword

        OUTPUT: the message corresponding to the received codeword

        EXAMPLES::

            sage: Fqm = GF(2^9)
            sage: Fq = GF(2^3)
            sage: C = codes.GabidulinCode(Fqm, 2, 2, Fq)
            sage: D = codes.decoders.GabidulinGaoDecoder(C)
            sage: E = codes.encoders.GabidulinPolynomialEvaluationEncoder(C)
            sage: S.<x> = Fqm['x', C.twisting_homomorphism()]
            sage: z9 = Fqm.gen()
            sage: p = (z9^6 + z9^4)*x + z9^2 + z9
            sage: codeword_vector = E.encode(p, "vector")
            sage: r = D.decode_to_message(codeword_vector)
            sage: r
            (z9^6 + z9^4)*x + z9^2 + z9
        """
        return self._decode_to_code_and_message(r)[1]

    def decoding_radius(self):
        """
        Return the decoding radius of the Gabidulin Gao Decoder.

        EXAMPLES::

            sage: Fqm = GF(5^20)
            sage: Fq = GF(5)
            sage: C = codes.GabidulinCode(Fqm, 20, 4, Fq)
            sage: D = codes.decoders.GabidulinGaoDecoder(C)
            sage: D.decoding_radius()
            8
        """
        return (self.code().minimum_distance() - 1) // 2

# ----------------------------- registration --------------------------------


GabidulinCode._registered_encoders["PolynomialEvaluation"] = GabidulinPolynomialEvaluationEncoder
GabidulinCode._registered_encoders["VectorEvaluation"] = GabidulinVectorEvaluationEncoder

GabidulinCode._registered_decoders["Gao"] = GabidulinGaoDecoder



# sage_setup: distribution = sagemath-schemes
# sage.doctest: needs sage.rings.number_field
r"""
The set `\mathbb{P}^1(K)` of cusps of a number field `K`

AUTHORS:

- Maite Aranes (2009): Initial version

EXAMPLES:

The space of cusps over a number field k::

    sage: x = polygen(ZZ, 'x')
    sage: k.<a> = NumberField(x^2 + 5)
    sage: kCusps = NFCusps(k); kCusps
    Set of all cusps of Number Field in a with defining polynomial x^2 + 5
    sage: kCusps is NFCusps(k)
    True

Define a cusp over a number field::

    sage: NFCusp(k, a, 2/(a+1))
    Cusp [a - 5: 2] of Number Field in a with defining polynomial x^2 + 5
    sage: kCusps((a,2))
    Cusp [a: 2] of Number Field in a with defining polynomial x^2 + 5
    sage: NFCusp(k,oo)
    Cusp Infinity of Number Field in a with defining polynomial x^2 + 5

Different operations with cusps over a number field::

    sage: alpha = NFCusp(k, 3, 1/a + 2); alpha
    Cusp [a + 10: 7] of Number Field in a with defining polynomial x^2 + 5
    sage: alpha.numerator()
    a + 10
    sage: alpha.denominator()
    7
    sage: alpha.ideal()
    Fractional ideal (7, a + 3)
    sage: M = alpha.ABmatrix(); M # random
    [a + 10, 2*a + 6, 7, a + 5]
    sage: NFCusp(k, oo).apply(M)
    Cusp [a + 10: 7] of Number Field in a with defining polynomial x^2 + 5

Check Gamma0(N)-equivalence of cusps::

    sage: N = k.ideal(3)
    sage: alpha = NFCusp(k, 3, a + 1)
    sage: beta = kCusps((2, a - 3))
    sage: alpha.is_Gamma0_equivalent(beta, N)
    True

Obtain transformation matrix for equivalent cusps::

    sage: t, M = alpha.is_Gamma0_equivalent(beta, N, Transformation=True)
    sage: M[2] in N
    True
    sage: M[0]*M[3] - M[1]*M[2] == 1
    True
    sage: alpha.apply(M) == beta
    True

List representatives for Gamma_0(N) - equivalence classes of cusps::

    sage: Gamma0_NFCusps(N)
    [Cusp [0: 1] of Number Field in a with defining polynomial x^2 + 5,
     Cusp [1: 3] of Number Field in a with defining polynomial x^2 + 5,
    ...]
"""
# ****************************************************************************
#       Copyright (C) 2009, Maite Aranes <M.T.Aranes@warwick.ac.uk>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from typing import Any

from sage.misc.cachefunc import cached_method, cached_function
from sage.structure.parent import Parent
from sage.structure.element import Element, InfinityElement
from sage.structure.richcmp import richcmp, rich_to_bool
from sage.structure.unique_representation import UniqueRepresentation


@cached_function
def list_of_representatives(N):
    """
    Return a list of ideals, coprime to the ideal ``N``, representatives of
    the ideal classes of the corresponding number field.

    .. NOTE::

        This list, used every time we check `\\Gamma_0(N)` - equivalence of
        cusps, is cached.

    INPUT:

    - ``N`` -- an ideal of a number field

    OUTPUT:

    A list of ideals coprime to the ideal ``N``, such that they are
    representatives of all the ideal classes of the number field.

    EXAMPLES::

        sage: from sage.modular.cusps_nf import list_of_representatives
        sage: x = polygen(ZZ, 'x')
        sage: k.<a> = NumberField(x^4 + 13*x^3 - 11)
        sage: N = k.ideal(713, a + 208)
        sage: L = list_of_representatives(N); L
        (Fractional ideal (1),
         Fractional ideal (47, a - 9),
         Fractional ideal (53, a - 16))
    """
    return NFCusps_ideal_reps_for_levelN(N)[0]


@cached_function
def NFCusps(number_field):
    r"""
    The set of cusps of a number field `K`, i.e. `\mathbb{P}^1(K)`.

    INPUT:

    - ``number_field`` -- a number field

    OUTPUT: the set of cusps over the given number field

    EXAMPLES::

        sage: x = polygen(ZZ, 'x')
        sage: k.<a> = NumberField(x^2 + 5)
        sage: kCusps = NFCusps(k); kCusps
        Set of all cusps of Number Field in a with defining polynomial x^2 + 5
        sage: kCusps is NFCusps(k)
        True

    Saving and loading works::

        sage: loads(kCusps.dumps()) == kCusps
        True
    """
    return NFCuspsSpace(number_field)


# *************************************************************************
#        NFCuspsSpace class                                               *
# *************************************************************************


class NFCuspsSpace(UniqueRepresentation, Parent):
    """
    The set of cusps of a number field. See ``NFCusps`` for full documentation.

    EXAMPLES::

        sage: x = polygen(ZZ, 'x')
        sage: k.<a> = NumberField(x^2 + 5)
        sage: kCusps = NFCusps(k); kCusps
        Set of all cusps of Number Field in a with defining polynomial x^2 + 5
    """
    def __init__(self, number_field):
        """
        See ``NFCusps`` for full documentation.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^3 + x^2 + 13)
            sage: kCusps = NFCusps(k); kCusps
            Set of all cusps of Number Field in a with defining polynomial x^3 + x^2 + 13
        """
        self.__number_field = number_field
        Parent.__init__(self, self)

    def __eq__(self, right):
        """
        Return equality only if right is the set of cusps for the same field.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 5)
            sage: L.<a> = NumberField(x^2 + 23)
            sage: kCusps = NFCusps(k); kCusps
            Set of all cusps of Number Field in a with defining polynomial x^2 + 5
            sage: LCusps = NFCusps(L); LCusps
            Set of all cusps of Number Field in a with defining polynomial x^2 + 23
            sage: kCusps == NFCusps(k)
            True
            sage: LCusps == NFCusps(L)
            True
            sage: LCusps == kCusps
            False
        """
        if not isinstance(right, NFCuspsSpace):
            return False
        return self.number_field() == right.number_field()

    def __ne__(self, right):
        """
        Check that ``self`` is not equal to ``right``.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 5)
            sage: L.<a> = NumberField(x^2 + 23)
            sage: kCusps = NFCusps(k); kCusps
            Set of all cusps of Number Field in a with defining polynomial x^2 + 5
            sage: LCusps = NFCusps(L); LCusps
            Set of all cusps of Number Field in a with defining polynomial x^2 + 23
            sage: kCusps != NFCusps(k)
            False
            sage: LCusps != NFCusps(L)
            False
            sage: LCusps != kCusps
            True
        """
        return not (self == right)

    def _repr_(self):
        """
        String representation of the set of cusps of a number field.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 2)
            sage: kCusps = NFCusps(k)
            sage: kCusps
            Set of all cusps of Number Field in a with defining polynomial x^2 + 2
            sage: kCusps._repr_()
            'Set of all cusps of Number Field in a with defining polynomial x^2 + 2'
            sage: kCusps.rename('Number Field Cusps'); kCusps
            Number Field Cusps
            sage: kCusps.rename(); kCusps
            Set of all cusps of Number Field in a with defining polynomial x^2 + 2
        """
        return "Set of all cusps of %s" % self.number_field()

    def _latex_(self):
        r"""
        Return latex representation of ``self``.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 5)
            sage: kCusps = NFCusps(k)
            sage: latex(kCusps) # indirect doctest
            \mathbf{P}^1(\Bold{Q}[a]/(a^{2} + 5))
        """
        return r"\mathbf{P}^1(%s)" % self.number_field()._latex_()

    def __call__(self, x):
        """
        Convert x into the set of cusps of a number field.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 5)
            sage: kCusps = NFCusps(k)
            sage: c = kCusps(a,2)
            Traceback (most recent call last):
            ...
            TypeError: ...__call__() takes 2 positional arguments but 3 were given

         ::

            sage: c = kCusps((a,2)); c
            Cusp [a: 2] of Number Field in a with defining polynomial x^2 + 5
            sage: kCusps(2/a)
            Cusp [-2*a: 5] of Number Field in a with defining polynomial x^2 + 5
            sage: kCusps(oo)
            Cusp Infinity of Number Field in a with defining polynomial x^2 + 5
        """
        return NFCusp(self.number_field(), x, parent=self)

    @cached_method
    def zero(self):
        """
        Return the zero cusp.

        .. NOTE::

            This method just exists to make some general algorithms work.
            It is not intended that the returned cusp is an additive
            neutral element.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 5)
            sage: kCusps = NFCusps(k)
            sage: kCusps.zero()
            Cusp [0: 1] of Number Field in a with defining polynomial x^2 + 5
        """
        return self(0)

    def number_field(self):
        """
        Return the number field that this set of cusps is attached to.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 1)
            sage: kCusps = NFCusps(k)
            sage: kCusps.number_field()
            Number Field in a with defining polynomial x^2 + 1
        """
        return self.__number_field

# *************************************************************************
#        NFCusp class                                                     *
# *************************************************************************


class NFCusp(Element):
    r"""
    Create a number field cusp, i.e., an element of `\mathbb{P}^1(k)`.

    A cusp on a number field is either an element of the field or infinity,
    i.e., an element of the projective line over the number field.  It is
    stored as a pair (a,b), where a, b are integral elements of the number
    field.

    INPUT:

    - ``number_field`` -- the number field over which the cusp is defined

    - ``a`` -- it can be a number field element (integral or not), or
      a number field cusp

    - ``b`` -- (optional) when present, it must be either Infinity or
      coercible to an element of the number field

    - ``lreps`` -- (optional) a list of chosen representatives for all the
      ideal classes of the field. When given, the representative of the cusp
      will be changed so its associated ideal is one of the ideals in the list.

    OUTPUT:

    ``[a: b]`` -- a number field cusp.

    EXAMPLES::

        sage: x = polygen(ZZ, 'x')
        sage: k.<a> = NumberField(x^2 + 5)
        sage: NFCusp(k, a, 2)
        Cusp [a: 2] of Number Field in a with defining polynomial x^2 + 5
        sage: NFCusp(k, (a,2))
        Cusp [a: 2] of Number Field in a with defining polynomial x^2 + 5
        sage: NFCusp(k, a, 2/(a+1))
        Cusp [a - 5: 2] of Number Field in a with defining polynomial x^2 + 5

    Cusp Infinity:

    ::

        sage: NFCusp(k, 0)
        Cusp [0: 1] of Number Field in a with defining polynomial x^2 + 5
        sage: NFCusp(k, oo)
        Cusp Infinity of Number Field in a with defining polynomial x^2 + 5
        sage: NFCusp(k, 3*a, oo)
        Cusp [0: 1] of Number Field in a with defining polynomial x^2 + 5
        sage: NFCusp(k, a + 5, 0)
        Cusp Infinity of Number Field in a with defining polynomial x^2 + 5

    Saving and loading works:

    ::

        sage: alpha = NFCusp(k, a, 2/(a+1))
        sage: loads(dumps(alpha))==alpha
        True

    Some tests:

    ::

        sage: I*I
        -1
        sage: NFCusp(k, I)
        Traceback (most recent call last):
        ...
        TypeError: unable to convert I to a cusp of the number field

    ::

        sage: NFCusp(k, oo, oo)
        Traceback (most recent call last):
        ...
        TypeError: unable to convert (+Infinity, +Infinity) to a cusp of the number field

    ::

        sage: NFCusp(k, 0, 0)
        Traceback (most recent call last):
        ...
        TypeError: unable to convert (0, 0) to a cusp of the number field

    ::

        sage: NFCusp(k, "a + 2", a)
        Cusp [-2*a + 5: 5] of Number Field in a with defining polynomial x^2 + 5

    ::

        sage: NFCusp(k, NFCusp(k, oo))
        Cusp Infinity of Number Field in a with defining polynomial x^2 + 5
        sage: c = NFCusp(k, 3, 2*a)
        sage: NFCusp(k, c, a + 1)
        Cusp [-a - 5: 20] of Number Field in a with defining polynomial x^2 + 5
        sage: L.<b> = NumberField(x^2 + 2)
        sage: NFCusp(L, c)
        Traceback (most recent call last):
        ...
        ValueError: Cannot coerce cusps from one field to another
    """
    def __init__(self, number_field, a, b=None, parent=None, lreps=None):
        """
        Constructor of number field cusps. See ``NFCusp`` for full
        documentation.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 1)
            sage: c = NFCusp(k, 3, a+1); c
            Cusp [3: a + 1] of Number Field in a with defining polynomial x^2 + 1
            sage: c.parent()
            Set of all cusps of Number Field in a with defining polynomial x^2 + 1
            sage: kCusps = NFCusps(k)
            sage: c.parent() is kCusps
            True
        """
        if parent is None:
            parent = NFCusps(number_field)
        Element.__init__(self, parent)
        R = number_field.maximal_order()
        if b is None:
            if not a:  # that is cusp "0"
                self.__a = R.zero()
                self.__b = R.one()
                return
            if isinstance(a, NFCusp):
                if a.parent() == parent:
                    self.__a = R(a.__a)
                    self.__b = R(a.__b)
                else:
                    raise ValueError("Cannot coerce cusps from one field to another")
            elif a in R:
                self.__a = R(a)
                self.__b = R.one()
            elif a in number_field:
                self.__b = R(a.denominator())
                self.__a = R(a * self.__b)
            elif isinstance(a, InfinityElement):
                self.__a = R.one()
                self.__b = R.zero()
            elif isinstance(a, int):
                self.__a = R(a)
                self.__b = R.one()
            elif isinstance(a, (tuple, list)):
                if len(a) != 2:
                    raise TypeError("unable to convert %r to a cusp \
                                      of the number field" % a)
                if a[1].is_zero():
                    self.__a = R.one()
                    self.__b = R.zero()
                elif a[0] in R and a[1] in R:
                    self.__a = R(a[0])
                    self.__b = R(a[1])
                elif isinstance(a[0], NFCusp):  # we know that a[1] is not zero
                    if a[1] == 1:
                        self.__a = a[0].__a
                        self.__b = a[0].__b
                    else:
                        r = a[0].__a / (a[0].__b * a[1])
                        self.__b = R(r.denominator())
                        self.__a = R(r * self.__b)
                else:
                    try:
                        r = number_field(a[0] / a[1])
                        self.__b = R(r.denominator())
                        self.__a = R(r * self.__b)
                    except (ValueError, TypeError):
                        raise TypeError(f"unable to convert {a} to a cusp "
                                        "of the number field")
            else:
                try:
                    r = number_field(a)
                    self.__b = R(r.denominator())
                    self.__a = R(r * self.__b)
                except (ValueError, TypeError):
                    raise TypeError("unable to convert %r to a cusp "
                                    "of the number field" % a)
        else:  # 'b' is given
            if isinstance(b, InfinityElement):
                if isinstance(a, InfinityElement) or (isinstance(a, NFCusp) and a.is_infinity()):
                    raise TypeError("unable to convert (%r, %r) "
                                    "to a cusp of the number field" % (a, b))
                self.__a = R.zero()
                self.__b = R.one()
                return
            elif not b:
                if not a:
                    raise TypeError("unable to convert (%r, %r) "
                                    "to a cusp of the number field" % (a, b))
                self.__a = R.one()
                self.__b = R.zero()
                return
            if not a:
                self.__a = R.zero()
                self.__b = R.one()
                return
            if (b in R or isinstance(b, int)) and (a in R or isinstance(a, int)):
                self.__a = R(a)
                self.__b = R(b)
            else:
                if a in R or a in number_field:
                    r = a / b
                elif isinstance(a, InfinityElement):
                    self.__a = R.one()
                    self.__b = R.zero()
                    return
                elif isinstance(a, NFCusp):
                    if a.is_infinity():
                        self.__a = R.one()
                        self.__b = R.zero()
                        return
                    r = a.__a / (a.__b * b)
                elif isinstance(a, int):
                    r = R(a) / b
                elif isinstance(a, (tuple, list)):
                    if len(a) != 2:
                        raise TypeError("unable to convert (%r, %r) \
                                          to a cusp of the number field" % (a, b))
                    r = R(a[0]) / (R(a[1]) * b)
                else:
                    try:
                        r = number_field(a) / b
                    except (ValueError, TypeError):
                        raise TypeError("unable to convert (%r, %r) \
                                          to a cusp of the number field" % (a, b))
                self.__b = R(r.denominator())
                self.__a = R(r * self.__b)
        if lreps is not None:
            # Changes the representative of the cusp so the ideal associated
            # to the cusp is one of the ideals of the given list lreps.
            # Note: the trivial class is always represented by (1).
            I = self.ideal()
            for J in lreps:
                if (J / I).is_principal():
                    newI = J
            l = (newI / I).gens_reduced()[0]
            self.__a = R(l * self.__a)
            self.__b = R(l * self.__b)

    def _repr_(self):
        """
        String representation of this cusp.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 1)
            sage: c = NFCusp(k, a, 2); c
            Cusp [a: 2] of Number Field in a with defining polynomial x^2 + 1
            sage: c._repr_()
            'Cusp [a: 2] of Number Field in a with defining polynomial x^2 + 1'
            sage: c.rename('[a:2](cusp of a number field)');c
            [a:2](cusp of a number field)
            sage: c.rename();c
            Cusp [a: 2] of Number Field in a with defining polynomial x^2 + 1
        """
        if self.__b.is_zero():
            return "Cusp Infinity of %s" % self.parent().number_field()
        else:
            return "Cusp [%s: %s] of %s" % (self.__a, self.__b,
                                            self.parent().number_field())

    def number_field(self):
        """
        Return the number field of definition of the cusp ``self``.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 2)
            sage: alpha = NFCusp(k, 1, a + 1)
            sage: alpha.number_field()
            Number Field in a with defining polynomial x^2 + 2
        """
        return self.parent().number_field()

    def is_infinity(self) -> bool:
        """
        Return ``True`` if this is the cusp infinity.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 1)
            sage: NFCusp(k, a, 2).is_infinity()
            False
            sage: NFCusp(k, 2, 0).is_infinity()
            True
            sage: NFCusp(k, oo).is_infinity()
            True
        """
        return self.__b == 0

    def numerator(self):
        """
        Return the numerator of the cusp ``self``.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 1)
            sage: c = NFCusp(k, a, 2)
            sage: c.numerator()
            a
            sage: d = NFCusp(k, 1, a)
            sage: d.numerator()
            1
            sage: NFCusp(k, oo).numerator()
            1
        """
        return self.__a

    def denominator(self):
        """
        Return the denominator of the cusp ``self``.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 1)
            sage: c = NFCusp(k, a, 2)
            sage: c.denominator()
            2
            sage: d = NFCusp(k, 1, a + 1);d
            Cusp [1: a + 1] of Number Field in a with defining polynomial x^2 + 1
            sage: d.denominator()
            a + 1
            sage: NFCusp(k, oo).denominator()
            0
        """
        return self.__b

    def _number_field_element_(self):
        """
        Coerce to an element of the number field.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 2)
            sage: NFCusp(k, a, 2)._number_field_element_()
            1/2*a
            sage: NFCusp(k, 1, a + 1)._number_field_element_()
            -1/3*a + 1/3
        """
        if self.__b.is_zero():
            raise TypeError("%s is not an element of %s" % (self,
                                                            self.number_field()))
        k = self.number_field()
        return k(self.__a / self.__b)

    def _ring_of_integers_element_(self):
        """
        Coerce to an element of the ring of integers of the number field.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 2)
            sage: NFCusp(k, a+1)._ring_of_integers_element_()
            a + 1
            sage: NFCusp(k, 1, a + 1)._ring_of_integers_element_()
            Traceback (most recent call last):
            ...
            TypeError: Cusp [1: a + 1] of Number Field in a with defining polynomial x^2 + 2 is not an integral element
        """
        if self.__b.is_one():
            return self.__a
        R = self.number_field().ring_of_integers()
        if self.__b.is_zero():
            raise TypeError("%s is not an element of %s" % (self, R))
        try:
            return R(self.__a / self.__b)
        except (ValueError, TypeError):
            raise TypeError("%s is not an integral element" % self)

    def _latex_(self):
        r"""
        Latex representation of this cusp.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 11)
            sage: latex(NFCusp(k, 3*a, a + 1)) # indirect doctest
            \[3 a: a + 1\]
            sage: latex(NFCusp(k, 3*a, a + 1)) == NFCusp(k, 3*a, a + 1)._latex_()
            True
            sage: latex(NFCusp(k, oo))
            \infty
        """
        if self.__b.is_zero():
            return "\\infty"
        else:
            return "\\[%s: %s\\]" % (self.__a._latex_(),
                                     self.__b._latex_())

    def _richcmp_(self, right, op):
        """
        Compare the cusps ``self`` and ``right``.

        Comparison is as for elements in the number field, except with
        the cusp oo which is greater than everything but itself.

        The ordering in comparison is only really meaningful for infinity.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^3 + x + 1)
            sage: kCusps = NFCusps(k)

        Comparing with infinity::

            sage: c = kCusps((a,2))
            sage: d = kCusps(oo)
            sage: c < d
            True
            sage: kCusps(oo) < d
            False

        Comparison as elements of the number field::

            sage: kCusps(2/3) < kCusps(5/2)
            False
            sage: k(2/3) < k(5/2)
            False
        """
        if self.__b.is_zero():
            if right.__b.is_zero():
                return rich_to_bool(op, 0)
            else:
                return rich_to_bool(op, 1)
        else:
            if right.__b.is_zero():
                return rich_to_bool(op, -1)
            else:
                return richcmp(self._number_field_element_(),
                               right._number_field_element_(), op)

    def __neg__(self):
        """
        The negative of this cusp.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 23)
            sage: c = NFCusp(k, a, a+1); c
            Cusp [a: a + 1] of Number Field in a with defining polynomial x^2 + 23
            sage: -c
            Cusp [-a: a + 1] of Number Field in a with defining polynomial x^2 + 23
        """
        return NFCusp(self.parent().number_field(), -self.__a, self.__b)

    def apply(self, g):
        """
        Return g(``self``), where ``g`` is a 2x2 matrix, which we view as a
        linear fractional transformation.

        INPUT:

        - ``g`` -- list of integral elements [a, b, c, d] that are the
          entries of a 2x2 matrix

        OUTPUT:

        A number field cusp, obtained by the action of ``g`` on the cusp
        ``self``.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 23)
            sage: beta = NFCusp(k, 0, 1)
            sage: beta.apply([0, -1, 1, 0])
            Cusp Infinity of Number Field in a with defining polynomial x^2 + 23
            sage: beta.apply([1, a, 0, 1])
            Cusp [a: 1] of Number Field in a with defining polynomial x^2 + 23
        """
        k = self.number_field()
        return NFCusp(k, g[0] * self.__a + g[1] * self.__b,
                      g[2] * self.__a + g[3] * self.__b)

    def ideal(self):
        """
        Return the ideal associated to the cusp ``self``.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^2 + 23)
            sage: alpha = NFCusp(k, 3, a-1)
            sage: alpha.ideal()
            Fractional ideal (3, 1/2*a - 1/2)
            sage: NFCusp(k, oo).ideal()
            Fractional ideal (1)
        """
        k = self.number_field()
        return k.ideal(self.__a, self.__b)

    def ABmatrix(self) -> list:
        """
        Return AB-matrix associated to the cusp ``self``.

        Given R a Dedekind domain and A, B ideals of R in inverse classes, an
        AB-matrix is a matrix realizing the isomorphism between R+R and A+B.
        An AB-matrix associated to a cusp [a1: a2] is an AB-matrix with A the
        ideal associated to the cusp (A=<a1, a2>) and first column given by
        the coefficients of the cusp.

        EXAMPLES:

        ::

            sage: x = polygen(ZZ, 'x')
            sage: k.<a> = NumberField(x^3 + 11)
            sage: alpha = NFCusp(k, oo)
            sage: alpha.ABmatrix()
            [1, 0, 0, 1]

        ::

            sage: alpha = NFCusp(k, 0)
            sage: alpha.ABmatrix()
            [0, -1, 1, 0]

        Note that the AB-matrix associated to a cusp is not unique, and the
        output of the ``ABmatrix`` function may change.

        ::

            sage: alpha = NFCusp(k, 3/2, a-1)
            sage: M = alpha.ABmatrix()
            sage: M # random
            [-a^2 - a - 1, -3*a - 7, 8, -2*a^2 - 3*a + 4]
            sage: M[0] == alpha.numerator() and M[2] == alpha.denominator()
            True

        An AB-matrix associated to a cusp alpha will send Infinity to alpha:

        ::

            sage: alpha = NFCusp(k, 3, a-1)
            sage: M = alpha.ABmatrix()
            sage: (k.ideal(M[1], M[3])*alpha.ideal()).is_principal()
            True
            sage: M[0] == alpha.numerator() and M[2] == alpha.denominator()
            True
            sage: NFCusp(k, oo).apply(M) == alpha
            True
        """
        k = self.number_field()
        A = self.ideal()

        if self.is_infinity():
            return [1, 0, 0, 1]
        if not self:
            return [0, -1, 1, 0]

        if A.is_principal():
            B = k.ideal(1)
        else:
            B = k.ideal(A.gens_reduced()[1]) / A
        assert (A * B).is_principal()

        a1 = self.__a
        a2 = self.__b

        g = (A * B).gens_reduced()[0]
        Ainv = A**(-1)
        A1 = a1 * Ainv
        A2 = a2 * Ainv
        r = A1.element_1_mod(A2)
        b1 = -(1 - r) / a2 * g
        b2 = (r / a1) * g
        return [a1, b1, a2, b2]

    def is_Gamma0_equivalent(self, other, N,
                             Transformation=False) -> bool | tuple[bool, Any]:
        r"""
        Check if cusps ``self`` and ``other`` are `\Gamma_0(N)`- equivalent.

        INPUT:

        - ``other`` -- a number field cusp or a list of two number field
          elements which define a cusp

        - ``N`` -- an ideal of the number field (level)

        OUTPUT: boolean; ``True`` if the cusps are equivalent

        - a transformation matrix -- (if ``Transformation=True``) a list of
          integral elements [a, b, c, d] which are the entries of a 2x2 matrix
          M in `\Gamma_0(N)` such that M * ``self`` = ``other`` if ``other``
          and ``self`` are `\Gamma_0(N)`- equivalent. If ``self`` and ``other``
          are not equivalent it returns zero.

        EXAMPLES:

        ::

            sage: x = polygen(ZZ, 'x')
            sage: K.<a> = NumberField(x^3 - 10)
            sage: N = K.ideal(a - 1)
            sage: alpha = NFCusp(K, 0)
            sage: beta = NFCusp(K, oo)
            sage: alpha.is_Gamma0_equivalent(beta, N)
            False
            sage: alpha.is_Gamma0_equivalent(beta, K.ideal(1))
            True
            sage: b, M = alpha.is_Gamma0_equivalent(beta, K.ideal(1),Transformation=True)
            sage: alpha.apply(M)
            Cusp Infinity of Number Field in a with defining polynomial x^3 - 10

        ::

            sage: k.<a> = NumberField(x^2 + 23)
            sage: N = k.ideal(3)
            sage: alpha1 = NFCusp(k, a+1, 4)
            sage: alpha2 = NFCusp(k, a-8, 29)
            sage: alpha1.is_Gamma0_equivalent(alpha2, N)
            True
            sage: b, M = alpha1.is_Gamma0_equivalent(alpha2, N, Transformation=True)
            sage: alpha1.apply(M) == alpha2
            True
            sage: M[2] in N
            True
        """
        k = self.number_field()
        other = NFCusp(k, other)
        if not (self.ideal() / other.ideal()).is_principal():
            if not Transformation:
                return False
            else:
                return False, 0

        reps = list_of_representatives(N)
        alpha1 = NFCusp(k, self, lreps=reps)
        alpha2 = NFCusp(k, other, lreps=reps)

        delta = k.ideal(alpha1.__b) + N
        if (k.ideal(alpha2.__b) + N) != delta:
            if not Transformation:
                return False
            else:
                return False, 0

        M1 = alpha1.ABmatrix()
        M2 = alpha2.ABmatrix()

        A = alpha1.ideal()
        B = k.ideal(M1[1], M1[3])

        ABdelta = A * B * delta * delta

        units = units_mod_ideal(ABdelta)
        for u in units:
            if (M2[2] * M1[3] - u * M1[2] * M2[3]) in ABdelta:
                if not Transformation:
                    return True
                else:
                    AuxCoeff = [1, 0, 0, 1]
                    Aux = M2[2] * M1[3] - u * M1[2] * M2[3]
                    if Aux in A * B * N:
                        if u != 1:
                            AuxCoeff[3] = u
                    else:
                        A1 = (A * B * N) / ABdelta
                        A2 = B * k.ideal(M1[2] * M2[2]) / (A * ABdelta)
                        f = A1.element_1_mod(A2)
                        w = ((1 - f) * Aux) / (M1[2] * M2[2])
                        AuxCoeff[3] = u
                        AuxCoeff[1] = w
                    from sage.matrix.constructor import Matrix
                    Maux = Matrix(k, 2, AuxCoeff)
                    M1inv = Matrix(k, 2, M1).inverse()
                    Mtrans = Matrix(k, 2, M2) * Maux * M1inv
                    assert Mtrans[1][0] in N
                    return True, Mtrans.list()
        if not Transformation:
            return False
        else:
            return False, 0

# *************************************************************************
#  Global functions:
#    - Gamma0_NFCusps --compute list of inequivalent cusps
#    Internal use only:
#    - number_of_Gamma0_NFCusps -- useful to test Gamma0_NFCusps
#    - NFCusps_ideal_reps_for_levelN -- lists of reps for ideal classes
#    - units_mod_ideal -- needed to check Gamma0(N)-equiv of cusps
# *************************************************************************


def Gamma0_NFCusps(N):
    r"""
    Return a list of inequivalent cusps for `\Gamma_0(N)`, i.e., a set of
    representatives for the orbits of ``self`` on `\mathbb{P}^1(k)`.

    INPUT:

    - ``N`` -- an integral ideal of the number field k (the level)

    OUTPUT: list of inequivalent number field cusps

    EXAMPLES::

        sage: x = polygen(ZZ, 'x')
        sage: k.<a> = NumberField(x^2 + 5)
        sage: N = k.ideal(3)
        sage: L = Gamma0_NFCusps(N)

    The cusps in the list are inequivalent::

        sage: any(L[i].is_Gamma0_equivalent(L[j], N)
        ....:     for i in range(len(L)) for j in range(len(L)) if i < j)
        False

    We test that we obtain the right number of orbits::

        sage: from sage.modular.cusps_nf import number_of_Gamma0_NFCusps
        sage: len(L) == number_of_Gamma0_NFCusps(N)
        True

    Another example::

        sage: x = polygen(ZZ, 'x')
        sage: k.<a> = NumberField(x^4 - x^3 -21*x^2 + 17*x + 133)
        sage: N = k.ideal(5)
        sage: from sage.modular.cusps_nf import number_of_Gamma0_NFCusps
        sage: len(Gamma0_NFCusps(N)) == number_of_Gamma0_NFCusps(N) # long time (over 1 sec)
        True
    """
    # We create L a list of three lists, which are different and each a list of
    # prime ideals, coprime to N, representing the ideal classes of k
    L = NFCusps_ideal_reps_for_levelN(N, nlists=3)
    Laux = L[1] + L[2]
    Lreps = list_of_representatives(N)
    Lcusps = []

    k = N.number_field()

    for A in L[0]:
        # find B in inverse class:
        if A.is_trivial():
            B = k.ideal(1)
            # B = k.unit_ideal() produces an error because we need fract ideal
            g = 1
        else:
            Lbs = [P for P in Laux if (P * A).is_principal()]
            B = Lbs[0]
            g = (A * B).gens_reduced()[0]

        # for every divisor of N we have to find cusps
        from sage.arith.misc import divisors
        for d in divisors(N):
            # find delta prime coprime to B in inverse class of d*A
            # by searching in our list of auxiliary prime ideals
            Lds = [P for P in Laux
                   if (P * d * A).is_principal() and P.is_coprime(B)]
            deltap = Lds[0]
            a = (deltap * d * A).gens_reduced()[0]
            I = d + N / d
            # special case: A=B=d=<1>:
            if a.is_one() and I.is_trivial():
                Lcusps.append(NFCusp(k, 0, 1, lreps=Lreps))
            else:
                u = k.unit_group().gens()
                for b in I.invertible_residues_mod(u):
                    # Note: if I trivial, invertible_residues_mod returns [1]
                    # lift b to (R/a)star
                    # we need the part of d which is coprime to I, call it M
                    M = d.prime_to_idealM_part(I)
                    deltAM = deltap * A * M
                    u = (B * deltAM).element_1_mod(I)
                    v = (I * B).element_1_mod(deltAM)
                    newb = u * b + v
                    # build AB-matrix:
                    # ----> extended gcd for k.ideal(a), k.ideal(newb)
                    Y = k.ideal(newb).element_1_mod(k.ideal(a))
                    # if xa + yb = 1, cusp = y*g /a
                    Lcusps.append(NFCusp(k, Y * g, a, lreps=Lreps))
    return Lcusps


def number_of_Gamma0_NFCusps(N):
    """
    Return the total number of orbits of cusps under the action of the
    congruence subgroup `\\Gamma_0(N)`.

    INPUT:

    - ``N`` -- a number field ideal

    OUTPUT: integer; the number of orbits of cusps under Gamma0(N)-action

    EXAMPLES::

        sage: x = polygen(ZZ, 'x')
        sage: k.<a> = NumberField(x^3 + 11)
        sage: N = k.ideal(2, a+1)
        sage: from sage.modular.cusps_nf import number_of_Gamma0_NFCusps
        sage: number_of_Gamma0_NFCusps(N)
        4
        sage: L = Gamma0_NFCusps(N)
        sage: len(L) == number_of_Gamma0_NFCusps(N)
        True
        sage: k.<a> = NumberField(x^2 + 7)
        sage: N = k.ideal(9)
        sage: number_of_Gamma0_NFCusps(N)
        6
        sage: N = k.ideal(a*9 + 7)
        sage: number_of_Gamma0_NFCusps(N)
        24
    """
    k = N.number_field()
    # The number of Gamma0(N)-sub-orbits for each Gamma-orbit:
    from sage.arith.misc import divisors
    Ugens = [k(u) for u in k.unit_group().gens()]
    s = sum([len((d + N / d).invertible_residues_mod(Ugens))
             for d in divisors(N)])
    # There are h Gamma-orbits, with h class number of underlying number field.
    return s * k.class_number()


def NFCusps_ideal_reps_for_levelN(N, nlists=1):
    """
    Return a list of lists (``nlists`` different lists) of prime ideals,
    coprime to ``N``, representing every ideal class of the number field.

    INPUT:

    - ``N`` -- number field ideal

    - ``nlists`` -- (default: 1) the number of lists of prime ideals
      we want

    OUTPUT:

    A list of lists of ideals representatives of the ideal classes, all coprime
    to ``N``, representing every ideal.

    EXAMPLES::

        sage: x = polygen(ZZ, 'x')
        sage: k.<a> = NumberField(x^3 + 11)
        sage: N = k.ideal(5, a + 1)
        sage: from sage.modular.cusps_nf import NFCusps_ideal_reps_for_levelN
        sage: NFCusps_ideal_reps_for_levelN(N)
        [(Fractional ideal (1), Fractional ideal (2, a + 1))]
        sage: L = NFCusps_ideal_reps_for_levelN(N, 3)
        sage: all(len(L[i]) == k.class_number() for i in range(len(L)))
        True

    ::

        sage: k.<a> = NumberField(x^4 - x^3 - 21*x^2 + 17*x + 133)
        sage: N = k.ideal(6)
        sage: from sage.modular.cusps_nf import NFCusps_ideal_reps_for_levelN
        sage: NFCusps_ideal_reps_for_levelN(N)
        [(Fractional ideal (1),
          Fractional ideal (67, -4/7*a^3 + 13/7*a^2 + 39/7*a - 43),
          Fractional ideal (127, -4/7*a^3 + 13/7*a^2 + 39/7*a - 42),
          Fractional ideal (157, -4/7*a^3 + 13/7*a^2 + 39/7*a + 48))]
        sage: L = NFCusps_ideal_reps_for_levelN(N, 5)
        sage: all(len(L[i]) == k.class_number() for i in range(len(L)))
        True
    """
    k = N.number_field()
    G = k.class_group()
    L = [[k.ideal(1)] for _ in range(nlists)]
    it = k.primes_of_degree_one_iter()
    for I in G.list():
        check = 0
        if not I.is_principal():
            Iinv = I.ideal()**(-1)
            while check < nlists:
                J = next(it)
                if (J * Iinv).is_principal() and J.is_coprime(N):
                    L[check].append(J)
                    check += 1
    return [tuple(l) for l in L]


def units_mod_ideal(I):
    """
    Return integral elements of the number field representing the images of
    the global units modulo the ideal ``I``.

    INPUT:

    - ``I`` -- number field ideal

    OUTPUT:

    A list of integral elements of the number field representing the images of
    the global units modulo the ideal ``I``. Elements of the list might be
    equivalent to each other mod ``I``.

    EXAMPLES::

        sage: from sage.modular.cusps_nf import units_mod_ideal
        sage: x = polygen(ZZ, 'x')
        sage: k.<a> = NumberField(x^2 + 1)
        sage: I = k.ideal(a + 1)
        sage: units_mod_ideal(I)
        [1]
        sage: I = k.ideal(3)
        sage: units_mod_ideal(I)
        [1, a, -1, -a]

    ::

        sage: from sage.modular.cusps_nf import units_mod_ideal
        sage: k.<a> = NumberField(x^3 + 11)
        sage: k.unit_group()
        Unit group with structure C2 x Z of
         Number Field in a with defining polynomial x^3 + 11
        sage: I = k.ideal(5, a + 1)
        sage: units_mod_ideal(I)
        [1,
        2*a^2 + 4*a - 1,
        ...]

    ::

        sage: from sage.modular.cusps_nf import units_mod_ideal
        sage: k.<a> = NumberField(x^4 - x^3 -21*x^2 + 17*x + 133)
        sage: k.unit_group()
        Unit group with structure C6 x Z of
         Number Field in a with defining polynomial x^4 - x^3 - 21*x^2 + 17*x + 133
        sage: I = k.ideal(3)
        sage: U = units_mod_ideal(I)
        sage: all(U[j].is_unit() and (U[j] not in I) for j in range(len(U)))
        True
    """
    k = I.number_field()
    Uk = k.unit_group()
    Istar = I.idealstar(2)
    ulist = Uk.gens_values()
    elist = [Istar(I.ideallog(u)).order() for u in ulist]

    from sage.misc.mrange import xmrange

    return [k.prod(u**e for u, e in zip(ulist, ei)) for ei in xmrange(elist)]



# sage_setup: distribution = sagemath-gsl

from sage.probability.random_variable import (
    DiscreteRandomVariable,
    DiscreteProbabilitySpace)

from sage.probability.probability_distribution import (
    RealDistribution,
    SphericalDistribution,
    GeneralDiscreteDistribution)



# sage_setup: distribution = sagemath-modules
r"""
Toric lattice elements

This module was designed as a part of the framework for toric varieties
(:mod:`~sage.schemes.toric.variety`,
:mod:`~sage.schemes.toric.fano_variety`).

AUTHORS:

- Andrey Novoseltsev (2010-05-27): initial version.

TESTS:

Let's create some lattices first::

    sage: N = ToricLattice(3)
    sage: M = N.dual()

Now we are ready to create elements of toric lattices::

    sage: n = N([1,2,3])
    sage: n
    N(1, 2, 3)
    sage: m = M(1,2,3)
    sage: m
    M(1, 2, 3)

Dual lattices can act on each other::

    sage: n * m
    14
    sage: m * n
    14

You can also add elements of the same lattice or scale them::

    sage: 2 * n
    N(2, 4, 6)
    sage: n * 2
    N(2, 4, 6)
    sage: n + n
    N(2, 4, 6)

However, you cannot "mix wrong lattices" in your expressions::

    sage: n + m
    Traceback (most recent call last):
    ...
    TypeError: unsupported operand parent(s) for +:
    '3-d lattice N' and '3-d lattice M'
    sage: n * n
    Traceback (most recent call last):
    ...
    TypeError: elements of the same toric lattice cannot be multiplied!
    sage: n == m
    False

Note that ``n`` and ``m`` are not equal to each other even though they are
both "just (1,2,3)." Moreover, you cannot easily convert elements between
toric lattices::

    sage: M(n)
    Traceback (most recent call last):
    ...
    TypeError: N(1, 2, 3) cannot be converted to 3-d lattice M!

If you really need to consider elements of one lattice as elements of another,
you can either use intermediate conversion to "just a vector"::

    sage: ZZ3 = ZZ^3
    sage: n_in_M = M(ZZ3(n))
    sage: n_in_M
    M(1, 2, 3)
    sage: n == n_in_M
    False
    sage: n_in_M == m
    True

Or you can create a homomorphism from one lattice to any other::

    sage: h = N.hom(identity_matrix(3), M)
    sage: h(n)
    M(1, 2, 3)
"""
# The "tutorial" above is a truncated version of one in toric_lattice.py.


# ***************************************************************************
#       Copyright (C) 2010 Andrey Novoseltsev <novoselt@gmail.com>
#       Copyright (C) 2010 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from sage.libs.gmp.mpz cimport *

from sage.modules.vector_integer_dense cimport Vector_integer_dense
from sage.structure.coerce_exceptions import CoercionException
from sage.structure.element cimport Vector
from sage.rings.integer cimport Integer
from sage.structure.richcmp cimport richcmp_not_equal


def is_ToricLatticeElement(x):
    r"""
    Check if ``x`` is an element of a toric lattice.

    INPUT:

    - ``x`` -- anything

    OUTPUT: ``True`` if ``x`` is an element of a toric lattice, ``False`` otherwise

    EXAMPLES::

        sage: from sage.geometry.toric_lattice_element import (
        ....:   is_ToricLatticeElement)
        sage: is_ToricLatticeElement(1)
        doctest:warning...
        DeprecationWarning: The function is_ToricLatticeElement is deprecated;
        use 'isinstance(..., ToricLatticeElement)' instead.
        See https://github.com/sagemath/sage/issues/38126 for details.
        False
        sage: e = ToricLattice(3).an_element()
        sage: e
        N(1, 0, 0)
        sage: is_ToricLatticeElement(e)
        True
    """
    from sage.misc.superseded import deprecation_cython
    deprecation_cython(38126,
                       "The function is_ToricLatticeElement is deprecated; "
                       "use 'isinstance(..., ToricLatticeElement)' instead.")
    return isinstance(x, ToricLatticeElement)


# Why do we need a special class:
# - customize output to include lattice name
# - prohibit operations mixing "wrong" lattices
cdef class ToricLatticeElement(Vector_integer_dense):
    r"""
    Create an element of a toric lattice.

    .. WARNING::

        You probably should not construct such elements explicitly.

    INPUT:

    - same as for
      :class:`~sage.modules.vector_integer_dense.Vector_integer_dense`.

    OUTPUT: element of a toric lattice

    TESTS::

        sage: N = ToricLattice(3)
        sage: from sage.geometry.toric_lattice_element import (
        ....:           ToricLatticeElement)
        sage: e = ToricLatticeElement(N, [1,2,3])
        sage: e
        N(1, 2, 3)
        sage: TestSuite(e).run()
    """
    def __richcmp__(self, right, op):
        r"""
        Compare ``self`` and ``right`` according to the operator ``op``.

        INPUT:

        - ``right`` -- another ToricLatticeElement

        OUTPUT: boolean

        First compare the ambient toric lattice, then compare the vectors.

        TESTS::

            sage: N = ToricLattice(3)
            sage: M = N.dual()
            sage: n = N(1,2,3)
            sage: m = M(1,2,3)
            sage: n == m
            False
            sage: n2 = N(1,2,3)
            sage: n == n2
            True
            sage: n is n2
            False
        """
        if not isinstance(right, ToricLatticeElement):
            return NotImplemented

        PL_ambient = self.parent().ambient_module()
        PR_ambient = right.parent().ambient_module()
        if PL_ambient != PR_ambient:
            return richcmp_not_equal(PL_ambient, PR_ambient, op)
        # Now use the real comparison of vectors
        return self._richcmp_(right, op)

    # For some reason, vectors work just fine without redefining this function
    # from the base class, but if it is not here, we get "unhashable type"...
    def __hash__(self):
        r"""
        Return the hash of ``self``.

        OUTPUT: integer

        TESTS::

            sage: N = ToricLattice(3)
            sage: n = N(1,2,3)
            sage: hash(n)
            Traceback (most recent call last):
            ...
            TypeError: mutable vectors are unhashable
            sage: n.set_immutable()
            sage: hash(n) == hash(n)
            True
        """
        return Vector_integer_dense.__hash__(self)

    cpdef _act_on_(self, other, bint self_on_left):
        """
        Act on ``other``.

        INPUT:

        - ``other`` -- :class:`ToricLatticeElement`

        OUTPUT:

        - integer, if ``other`` is an element of the dual lattice of ``self``;

        - ``CoercionException`` is raised if ``other`` is an element of
          an incompatible toric lattice;

        - standard output for ``self`` acting as an integral vector on
          ``other`` if the latter one is not an element of a toric lattice.

        TESTS::

            sage: N = ToricLattice(3)
            sage: M = N.dual()
            sage: n = N(1,2,3)
            sage: m = M(1,2,3)
            sage: n * m # indirect doctest
            14

        Now we test behaviour with other types::

            sage: v = vector([1, 2, 3])
            sage: v * n == n * v
            True
            sage: v = vector([1, 1/2, 3/4])
            sage: v * n == n * v
            True
            sage: A = matrix(3, range(9))
            sage: A * n
            (8, 26, 44)
            sage: n * A
            (24, 30, 36)
            sage: B = A / 3
            sage: B * n
            (8/3, 26/3, 44/3)
            sage: n * B
            (8, 10, 12)
        """
        Ns = self.parent()
        # We try to deal only with the case of two lattice elements...
        if isinstance(other, ToricLatticeElement):
            if other.parent().ambient_module() is Ns.ambient_module().dual():
                # Our own _dot_product_ is disabled
                return Vector_integer_dense._dot_product_(self, other)
            raise CoercionException("only elements of dual toric lattices "
                                    "can act on each other!")
        # ... however we also need to treat the case when other is an integral
        # vector, since otherwise it will be coerced to the parent of self and
        # then the dot product will be called for elements of the same lattice
        if isinstance(other, Vector_integer_dense):
            return Vector_integer_dense._dot_product_(self, other)
        # We also allow action on elements of lattice quotients
        try:
            lift = other.lift()
            if isinstance(lift, ToricLatticeElement):
                if other.parent().W().is_submodule(Ns.dual().W()):
                    return Vector_integer_dense._dot_product_(self, lift)
                raise CoercionException("only elements of dual toric lattices "
                                        "can act on each other!")
        except AttributeError:  # No lift
            pass
        # Now let the standard framework work...
        return Vector_integer_dense._act_on_(self, other, self_on_left)

    # We need to override this function to prohibit default behaviour.
    # It seems to be called when right is in the same lattice as self, which
    # is wrong from our point of view.
    cpdef _dot_product_(self, Vector right):
        """
        Raise a :exc:`TypeError` exception.

        Dot product is not defined on toric lattices (there are actions of
        dual lattices on each other instead).

        INPUT:

        - ``right`` -- vector

        OUTPUT: :exc:`TypeError` exception is raised

        TESTS::

            sage: N = ToricLattice(3)
            sage: M = N.dual()
            sage: n = N(1,2,3)
            sage: m = M(1,2,3)
            sage: n * n # indirect doctest
            Traceback (most recent call last):
            ...
            TypeError: elements of the same
            toric lattice cannot be multiplied!
        """
        raise TypeError("elements of the same toric lattice cannot be "
                        "multiplied!")

    def _latex_(self):
        r"""
        Return a LaTeX representation of ``self``.

        OUTPUT: string

        TESTS::

            sage: Ld = ToricLattice(3, "L").dual()
            sage: e = Ld(1,2,3)
            sage: e._latex_()
            '\\left(1,\\,2,\\,3\\right)_{L^*}'
        """
        return "%s_{%s}" % (super()._latex_(),
                            self.parent().ambient_module()._latex_name)

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        OUTPUT: string

        TESTS::

            sage: Ld = ToricLattice(3, "L").dual()
            sage: e = Ld(1,2,3)
            sage: e._repr_()
            'L*(1, 2, 3)'
        """
        return (self.parent().ambient_module()._name
                + super()._repr_())

    def __reduce__(self):
        """
        Override the base ``__reduce__`` to correctly pickle/unpickle elements.

        EXAMPLES::

            sage: N = ToricLattice(3)
            sage: loads(dumps(N(1,2,3)))
            N(1, 2, 3)
        """
        return (unpickle_v1, (self._parent, self.list(), self._degree,
                              not self._is_immutable))

    def plot(self, **options):
        r"""
        Plot ``self``.

        INPUT:

        - any options for toric plots (see :func:`toric_plotter.options
          <sage.geometry.toric_plotter.options>`), none are mandatory.

        OUTPUT: a plot

        EXAMPLES::

            sage: N = ToricLattice(3)
            sage: n = N(1,2,3)
            sage: n.plot()                                                              # needs sage.geometry.polyhedron sage.plot
            Graphics3d Object
        """
        from sage.geometry.toric_plotter import ToricPlotter
        tp = ToricPlotter(options, self.parent().degree())
        tp.adjust_options()
        return tp.plot_points([self])


def unpickle_v1(parent, entries, degree, is_mutable):
    """
    Unpickle a :class:`ToricLatticeElement`.

    INPUT:

    - ``parent`` -- the parent toric lattice

    - ``entries`` -- list; the coordinates of the lattice point

    - ``degree`` -- integer; the dimension of the toric lattice

    - ``is_mutable`` -- boolean; whether the lattice element is mutable

    OUTPUT: the :class:`ToricLatticeElement` determined by the input data

    EXAMPLES::

        sage: N = ToricLattice(3, "lattice")
        sage: loads(dumps(N(1,2,3)))   # indirect test
        lattice(1, 2, 3)
        sage: from sage.geometry.toric_lattice_element import unpickle_v1
        sage: unpickle_v1(N,[1,2,3],3,True)
        lattice(1, 2, 3)
    """
    cdef ToricLatticeElement v
    v = ToricLatticeElement.__new__(ToricLatticeElement)
    v._init(degree, parent)
    cdef Integer z
    cdef Py_ssize_t i
    for i in range(degree):
        z = Integer(entries[i])
        mpz_set(v._entries[i], z.value)
    v._is_immutable = not is_mutable
    return v



# sage_setup: distribution = sagemath-schemes
# sage.doctest: needs sage.libs.flint
r"""
Hecke operators and morphisms between modular abelian varieties

Sage can compute with Hecke operators on modular abelian varieties.
A Hecke operator is defined by given a modular abelian variety and
an index. Given a Hecke operator, Sage can compute the
characteristic polynomial, and the action of the Hecke operator on
various homology groups.

AUTHORS:

- William Stein (2007-03)

- Craig Citro (2008-03)

EXAMPLES::

    sage: A = J0(54)
    sage: t5 = A.hecke_operator(5); t5
    Hecke operator T_5 on Abelian variety J0(54) of dimension 4
    sage: t5.charpoly().factor()
    (x - 3) * (x + 3) * x^2
    sage: B = A.new_subvariety(); B
    Abelian subvariety of dimension 2 of J0(54)
    sage: t5 = B.hecke_operator(5); t5
    Hecke operator T_5 on Abelian subvariety of dimension 2 of J0(54)
    sage: t5.charpoly().factor()
    (x - 3) * (x + 3)
    sage: t5.action_on_homology().matrix()
    [ 0  3  3 -3]
    [-3  3  3  0]
    [ 3  3  0 -3]
    [-3  6  3 -3]
"""

###########################################################################
#       Copyright (C) 2007 William Stein <wstein@gmail.com>               #
#  Distributed under the terms of the GNU General Public License (GPL)    #
#                  http://www.gnu.org/licenses/                           #
###########################################################################

from sage.categories.morphism import Morphism as base_Morphism
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ

import sage.modules.matrix_morphism
import sage.matrix.matrix_space as matrix_space

from .finite_subgroup import TorsionPoint


class Morphism_abstract(sage.modules.matrix_morphism.MatrixMorphism_abstract):
    """
    A morphism between modular abelian varieties.

    EXAMPLES::

        sage: t = J0(11).hecke_operator(2)
        sage: from sage.modular.abvar.morphism import Morphism
        sage: isinstance(t, Morphism)
        True
    """

    def _repr_(self):
        r"""
        Return string representation of this morphism.

        EXAMPLES::

            sage: t = J0(11).hecke_operator(2)
            sage: sage.modular.abvar.morphism.Morphism_abstract._repr_(t)
            'Abelian variety endomorphism of Abelian variety J0(11) of dimension 1'
            sage: J0(42).projection(J0(42)[0])._repr_()
            'Abelian variety morphism:\n  From: Abelian variety J0(42) of dimension 5\n  To:   Simple abelian subvariety 14a(1,42) of dimension 1 of J0(42)'
        """
        return base_Morphism._repr_(self)

    def _repr_type(self):
        """
        Return type of morphism.

        EXAMPLES::

            sage: t = J0(11).hecke_operator(2)
            sage: sage.modular.abvar.morphism.Morphism_abstract._repr_type(t)
            'Abelian variety'
        """
        return "Abelian variety"

    def complementary_isogeny(self):
        """
        Return the complementary isogeny of ``self``.

        EXAMPLES::

            sage: J = J0(43)
            sage: A = J[1]
            sage: T5 = A.hecke_operator(5)
            sage: T5.is_isogeny()
            True
            sage: T5.complementary_isogeny()
            Abelian variety endomorphism of Simple abelian subvariety 43b(1,43) of dimension 2 of J0(43)
            sage: (T5.complementary_isogeny() * T5).matrix()
            [2 0 0 0]
            [0 2 0 0]
            [0 0 2 0]
            [0 0 0 2]
        """
        if not self.is_isogeny():
            raise ValueError("self is not an isogeny")
        M = self.matrix()
        try:
            iM, denom = M._invert_iml()
        except (AttributeError, ImportError):
            iM = M.matrix_over_field().inverse()
            iM, denom = iM._clear_denom()
        return Morphism(self.parent().reversed(), iM)

    def is_isogeny(self) -> bool:
        """
        Return ``True`` if this morphism is an isogeny of abelian varieties.

        EXAMPLES::

            sage: J = J0(39)
            sage: Id = J.hecke_operator(1)
            sage: Id.is_isogeny()
            True
            sage: J.hecke_operator(19).is_isogeny()
            False
        """
        M = self.matrix()
        return M.nrows() == M.ncols() == M.rank()

    def cokernel(self):
        """
        Return the cokernel of ``self``.

        OUTPUT:

        - ``A`` -- an abelian variety (the cokernel)

        - ``phi`` -- a quotient map from ``self.codomain()`` to the
          cokernel of self

        EXAMPLES::

            sage: t = J0(33).hecke_operator(2)
            sage: (t-1).cokernel()
            (Abelian subvariety of dimension 1 of J0(33),
             Abelian variety morphism:
              From: Abelian variety J0(33) of dimension 3
              To:   Abelian subvariety of dimension 1 of J0(33))

        Projection will always have cokernel zero.

        ::

            sage: J0(37).projection(J0(37)[0]).cokernel()
            (Simple abelian subvariety of dimension 0 of J0(37),
             Abelian variety morphism:
              From: Simple abelian subvariety 37a(1,37) of dimension 1 of J0(37)
              To:   Simple abelian subvariety of dimension 0 of J0(37))

        Here we have a nontrivial cokernel of a Hecke operator, as the
        T_2-eigenvalue for the newform 37b is 0.

        ::

            sage: J0(37).hecke_operator(2).cokernel()
            (Abelian subvariety of dimension 1 of J0(37),
             Abelian variety morphism:
              From: Abelian variety J0(37) of dimension 2
              To:   Abelian subvariety of dimension 1 of J0(37))
            sage: AbelianVariety('37b').newform().q_expansion(5)
            q + q^3 - 2*q^4 + O(q^5)
        """
        try:
            return self.__cokernel
        except AttributeError:
            I = self.image()
            C = self.codomain().quotient(I)
            self.__cokernel = C
            return C

    def kernel(self):
        """
        Return the kernel of this morphism.

        OUTPUT:

        - ``G`` -- a finite group

        - ``A`` -- an abelian variety (identity component of the kernel)

        EXAMPLES: We compute the kernel of a projection map. Notice that
        the kernel has a nontrivial abelian variety part.

        ::

            sage: A, B, C = J0(33)
            sage: pi = J0(33).projection(B)
            sage: pi.kernel()
            (Finite subgroup with invariants [20] over QQbar of Abelian variety J0(33) of dimension 3,
             Abelian subvariety of dimension 2 of J0(33))

        We compute the kernels of some Hecke operators::

            sage: t2 = J0(33).hecke_operator(2)
            sage: t2
            Hecke operator T_2 on Abelian variety J0(33) of dimension 3
            sage: t2.kernel()
            (Finite subgroup with invariants [2, 2, 2, 2] over QQ of Abelian variety J0(33) of dimension 3,
             Abelian subvariety of dimension 0 of J0(33))
            sage: t3 = J0(33).hecke_operator(3)
            sage: t3.kernel()
            (Finite subgroup with invariants [3, 3] over QQ of Abelian variety J0(33) of dimension 3,
             Abelian subvariety of dimension 0 of J0(33))
        """
        A = self.matrix()
        L = A.image().change_ring(ZZ)
        # Saturate the image of the matrix corresponding to self.
        Lsat = L.saturation()
        # Now find a matrix whose rows map exactly onto the
        # saturation of L.
        X = A.solve_left(Lsat.basis_matrix())
        D = self.domain()
        V = (A.kernel().basis_matrix() * D.vector_space().basis_matrix()).row_module()
        Lambda = V.intersection(D._ambient_lattice())
        from .abvar import ModularAbelianVariety
        abvar = ModularAbelianVariety(D.groups(), Lambda, D.base_ring())

        if Lambda.rank() == 0:
            field_of_definition = QQ
        else:
            field_of_definition = None

        lattice = (X * self.domain().lattice().basis_matrix()).row_module(ZZ)

        K = D.finite_subgroup(lattice, field_of_definition=field_of_definition)

        return K, abvar

    def factor_out_component_group(self):
        r"""
        View ``self`` as a morphism `f:A \to B`. Then `\ker(f)`
        is an extension of an abelian variety `C` by a finite
        component group `G`. This function constructs a morphism
        `g` with domain `A` and codomain Q isogenous to
        `C` such that `\ker(g)` is equal to `C`.

        OUTPUT: a morphism

        EXAMPLES::

            sage: A,B,C = J0(33)
            sage: pi = J0(33).projection(A)
            sage: pi.kernel()
            (Finite subgroup with invariants [5] over QQbar of Abelian variety J0(33) of dimension 3,
             Abelian subvariety of dimension 2 of J0(33))
            sage: psi = pi.factor_out_component_group()
            sage: psi.kernel()
            (Finite subgroup with invariants [] over QQbar of Abelian variety J0(33) of dimension 3,
             Abelian subvariety of dimension 2 of J0(33))

        ALGORITHM: We compute a subgroup `G` of `B` so that
        the composition `h: A\to B \to B/G` has kernel that
        contains `A[n]` and component group isomorphic to
        `(\ZZ/n\ZZ)^{2d}`, where `d` is the
        dimension of `A`. Then `h` factors through
        multiplication by `n`, so there is a morphism
        `g: A\to B/G` such that `g \circ [n] = h`. Then
        `g` is the desired morphism. We give more details below
        about how to transform this into linear algebra.
        """
        try:
            return self.__factor_out
        except AttributeError:
            A = self.matrix()
            L = A.image()
            # Saturate the image of the matrix corresponding to self.
            Lsat = L.saturation()
            if L == Lsat:  # easy case
                self.__factor_out = self
                return self
            # Now find a matrix whose rows map exactly onto the
            # saturation of L.
            X = A.solve_left(Lsat.basis_matrix())

            # Find an integer n such that 1/n times the lattice Lambda of A
            # contains the row span of X.
            n = X.denominator()

            # Let M be the lattice of the codomain B of self.
            # Now 1/n * Lambda contains Lambda and maps
            # via the matrix of self to a lattice L' that
            # contains Lsat.   Consider the lattice
            #                R = M + L'.
            # This is a lattice that contains the lattice M of B.
            # Also 1/n*Lambda maps exactly to L' in R.
            # We have
            #     R/L' = (M+L')/L' = M/(L'/\M) = M/Lsat
            # which is torsion free!

            Q = self.codomain()
            M = Q.lattice()
            one_over_n = ZZ.one() / n
            Lprime = (one_over_n * self.matrix() * M.basis_matrix()).row_module(ZZ)

            # This R is a lattice in the ambient space for B.
            R = Lprime + M

            from .abvar import ModularAbelianVariety
            C = ModularAbelianVariety(Q.groups(), R, Q.base_field())

            # We have to change the basis of the representation of A
            # to the basis for R instead of the basis for M.  Each row
            # of A is written in terms of M, but needs to be in terms
            # of R's basis, which contains M with finite index.
            change_basis_from_M_to_R = R.basis_matrix().solve_left(M.basis_matrix())
            matrix = one_over_n * A * change_basis_from_M_to_R

            # Finally
            g = Morphism(self.domain().Hom(C), matrix)
            self.__factor_out = g
            return g

    def image(self):
        """
        Return the image of this morphism.

        OUTPUT: an abelian variety

        EXAMPLES: We compute the image of projection onto a factor of
        `J_0(33)`::

            sage: A,B,C = J0(33)
            sage: A
            Simple abelian subvariety 11a(1,33) of dimension 1 of J0(33)
            sage: f = J0(33).projection(A)
            sage: f.image()
            Abelian subvariety of dimension 1 of J0(33)
            sage: f.image() == A
            True

        We compute the image of a Hecke operator::

            sage: t2 = J0(33).hecke_operator(2); t2.fcp()
            (x - 1) * (x + 2)^2
            sage: phi = t2 + 2
            sage: phi.image()
            Abelian subvariety of dimension 1 of J0(33)

        The sum of the image and the kernel is the whole space::

            sage: phi.kernel()[1] + phi.image() == J0(33)
            True
        """
        return self(self.domain())

    def __call__(self, X):
        """
        INPUT:

        - ``X`` -- abelian variety, finite group, or torsion element

        OUTPUT: abelian variety, finite group, torsion element

        EXAMPLES: We apply morphisms to elements::

            sage: t2 = J0(33).hecke_operator(2)
            sage: G  = J0(33).torsion_subgroup(2); G
            Finite subgroup with invariants [2, 2, 2, 2, 2, 2] over QQ of Abelian variety J0(33) of dimension 3
            sage: t2(G.0)
            [(-1/2, 0, 1/2, -1/2, 1/2, -1/2)]
            sage: t2(G.0) in G
            True
            sage: t2(G.1)
            [(0, -1, 1/2, 0, 1/2, -1/2)]
            sage: t2(G.2)
            [(0, 0, 0, 0, 0, 0)]
            sage: K = t2.kernel()[0]; K
            Finite subgroup with invariants [2, 2, 2, 2] over QQ of Abelian variety J0(33) of dimension 3
            sage: t2(K.0)
            [(0, 0, 0, 0, 0, 0)]

        We apply morphisms to subgroups::

            sage: t2 = J0(33).hecke_operator(2)
            sage: G  = J0(33).torsion_subgroup(2); G
            Finite subgroup with invariants [2, 2, 2, 2, 2, 2] over QQ of Abelian variety J0(33) of dimension 3
            sage: t2(G)
            Finite subgroup with invariants [2, 2] over QQ of Abelian variety J0(33) of dimension 3
            sage: t2.fcp()
            (x - 1) * (x + 2)^2

        We apply morphisms to abelian subvarieties::

            sage: E11a0, E11a1, B = J0(33)
            sage: t2 = J0(33).hecke_operator(2)
            sage: t3 = J0(33).hecke_operator(3)
            sage: E11a0
            Simple abelian subvariety 11a(1,33) of dimension 1 of J0(33)
            sage: t3(E11a0)
            Abelian subvariety of dimension 1 of J0(33)
            sage: t3(E11a0).decomposition()
            [Simple abelian subvariety 11a(3,33) of dimension 1 of J0(33)]
            sage: t3(E11a0) == E11a1
            True
            sage: t2(E11a0) == E11a0
            True
            sage: t3(E11a0) == E11a0
            False
            sage: t3(E11a0 + E11a1) == E11a0 + E11a1
            True

        We apply some Hecke operators to the cuspidal subgroup and split it
        up::

            sage: C = J0(33).cuspidal_subgroup(); C
            Finite subgroup with invariants [10, 10] over QQ of Abelian variety J0(33) of dimension 3
            sage: t2 = J0(33).hecke_operator(2); t2.fcp()
            (x - 1) * (x + 2)^2
            sage: (t2 - 1)(C)
            Finite subgroup with invariants [5, 5] over QQ of Abelian variety J0(33) of dimension 3
            sage: (t2 + 2)(C)
            Finite subgroup with invariants [2, 2] over QQ of Abelian variety J0(33) of dimension 3

        Same but on a simple new factor::

            sage: C = J0(33)[2].cuspidal_subgroup(); C
            Finite subgroup with invariants [2, 2] over QQ of Simple abelian subvariety 33a(1,33) of dimension 1 of J0(33)
            sage: t2 = J0(33)[2].hecke_operator(2); t2.fcp()
            x - 1
            sage: t2(C)
            Finite subgroup with invariants [2, 2] over QQ of Simple abelian subvariety 33a(1,33) of dimension 1 of J0(33)
        """
        from .abvar import ModularAbelianVariety_abstract
        from .finite_subgroup import FiniteSubgroup
        if isinstance(X, TorsionPoint):
            return self._image_of_element(X)
        elif isinstance(X, ModularAbelianVariety_abstract):
            return self._image_of_abvar(X)
        elif isinstance(X, FiniteSubgroup):
            return self._image_of_finite_subgroup(X)
        else:
            raise TypeError("X must be an abelian variety or finite subgroup")

    def _image_of_element(self, x):
        """
        Return the image of the torsion point `x` under this
        morphism.

        The parent of the image element is always the group of all torsion
        elements of the abelian variety.

        INPUT:

        - ``x`` -- a torsion point on an abelian variety

        OUTPUT: a torsion point

        EXAMPLES::

            sage: A = J0(11); t = A.hecke_operator(2)
            sage: t.matrix()
            [-2  0]
            [ 0 -2]
            sage: P = A.cuspidal_subgroup().0; P
            [(0, 1/5)]
            sage: t._image_of_element(P)
            [(0, -2/5)]
            sage: -2*P
            [(0, -2/5)]

        ::

            sage: J = J0(37) ; phi = J._isogeny_to_product_of_simples()
            sage: phi._image_of_element(J.torsion_subgroup(5).gens()[0])
            [(1/5, -1/5, -1/5, 1/5, 1/5, 1/5, 1/5, -1/5)]

        ::

            sage: K = J[0].intersection(J[1])[0]
            sage: K.list()
            [[(0, 0, 0, 0)],
             [(1/2, -1/2, 1/2, 0)],
             [(0, 0, 1, -1/2)],
             [(1/2, -1/2, 3/2, -1/2)]]
            sage: [ phi.restrict_domain(J[0])._image_of_element(k) for k in K ]
            [[(0, 0, 0, 0, 0, 0, 0, 0)],
             [(0, 0, 0, 0, 0, 0, 0, 0)],
             [(0, 0, 0, 0, 0, 0, 0, 0)],
             [(0, 0, 0, 0, 0, 0, 0, 0)]]
        """
        v = x._relative_element() * self.matrix() * self.codomain().lattice().basis_matrix()
        T = self.codomain().qbar_torsion_subgroup()
        return T(v)

    def _image_of_finite_subgroup(self, G):
        """
        Return the image of the finite group `G` under ``self``.

        INPUT:

        - ``G`` -- a finite subgroup of the domain of ``self``

        OUTPUT: a finite subgroup of the codomain

        EXAMPLES::

            sage: J = J0(33); A = J[0]; B = J[1]
            sage: C = A.intersection(B)[0]; C
            Finite subgroup with invariants [5] over QQ of Simple abelian subvariety 11a(1,33) of dimension 1 of J0(33)
            sage: t = J.hecke_operator(3)
            sage: D = t(C); D
            Finite subgroup with invariants [5] over QQ of Abelian variety J0(33) of dimension 3
            sage: D == C
            True

        Or we directly test this function::

            sage: D = t._image_of_finite_subgroup(C); D
            Finite subgroup with invariants [5] over QQ of Abelian variety J0(33) of dimension 3
            sage: phi = J0(11).degeneracy_map(22,2)
            sage: J0(11).rational_torsion_subgroup().order()
            5
            sage: phi._image_of_finite_subgroup(J0(11).rational_torsion_subgroup())
            Finite subgroup with invariants [5] over QQ of Abelian variety J0(22) of dimension 2
        """
        B = G._relative_basis_matrix() * self.restrict_domain(G.abelian_variety()).matrix() * self.codomain().lattice().basis_matrix()
        lattice = B.row_module(ZZ)
        return self.codomain().finite_subgroup(lattice,
            field_of_definition=G.field_of_definition())

    def _image_of_abvar(self, A):
        """
        Compute the image of the abelian variety `A` under this
        morphism.

        INPUT:

        - ``A`` -- an abelian variety

        OUTPUT an abelian variety

        EXAMPLES::

            sage: t = J0(33).hecke_operator(2)
            sage: t._image_of_abvar(J0(33).new_subvariety())
            Abelian subvariety of dimension 1 of J0(33)

        ::

            sage: t = J0(33).hecke_operator(3)
            sage: A = J0(33)[0]
            sage: B = t._image_of_abvar(A); B
            Abelian subvariety of dimension 1 of J0(33)
            sage: B == A
            False
            sage: A + B == J0(33).old_subvariety()
            True

        ::

             sage: J = J0(37) ; A, B = J.decomposition()
            sage: J.projection(A)._image_of_abvar(A)
            Abelian subvariety of dimension 1 of J0(37)
            sage: J.projection(A)._image_of_abvar(B)
            Abelian subvariety of dimension 0 of J0(37)
            sage: J.projection(B)._image_of_abvar(A)
            Abelian subvariety of dimension 0 of J0(37)
            sage: J.projection(B)._image_of_abvar(B)
            Abelian subvariety of dimension 1 of J0(37)
            sage: J.projection(B)._image_of_abvar(J)
            Abelian subvariety of dimension 1 of J0(37)
        """
        from .abvar import ModularAbelianVariety
        D = self.domain()
        C = self.codomain()
        if A is D:
            B = self.matrix()
        else:
            if not A.is_subvariety(D):
                raise ValueError("A must be an abelian subvariety of self.")
            # Write the vector space corresponding to A in terms of self's
            # vector space, then take the image under self.
            B = D.vector_space().coordinate_module(A.vector_space()).basis_matrix() * self.matrix()

        V = (B * C.vector_space().basis_matrix()).row_module(QQ)

        lattice = V.intersection(C.lattice())
        base_field = C.base_field()
        return ModularAbelianVariety(C.groups(), lattice, base_field)


class Morphism(Morphism_abstract, sage.modules.matrix_morphism.MatrixMorphism):

    def restrict_domain(self, sub):
        """
        Restrict ``self`` to the subvariety sub of ``self.domain()``.

        EXAMPLES::

            sage: J = J0(37) ; A, B = J.decomposition()
            sage: A.lattice().matrix()
            [ 1 -1  1  0]
            [ 0  0  2 -1]
            sage: B.lattice().matrix()
            [1 1 1 0]
            [0 0 0 1]
            sage: T = J.hecke_operator(2) ; T.matrix()
            [-1  1  1 -1]
            [ 1 -1  1  0]
            [ 0  0 -2  1]
            [ 0  0  0  0]
            sage: T.restrict_domain(A)
            Abelian variety morphism:
              From: Simple abelian subvariety 37a(1,37) of dimension 1 of J0(37)
              To:   Abelian variety J0(37) of dimension 2
            sage: T.restrict_domain(A).matrix()
            [-2  2 -2  0]
            [ 0  0 -4  2]
            sage: T.restrict_domain(B)
            Abelian variety morphism:
              From: Simple abelian subvariety 37b(1,37) of dimension 1 of J0(37)
              To:   Abelian variety J0(37) of dimension 2
            sage: T.restrict_domain(B).matrix()
            [0 0 0 0]
            [0 0 0 0]
        """
        if not sub.is_subvariety(self.domain()):
            raise ValueError("sub must be a subvariety of self.domain()")

        if sub == self.domain():
            return self

        L = self.domain().lattice()
        B = sub.lattice().basis()
        ims = sum(((L(b) * self.matrix()).list() for b in B), [])
        MS = matrix_space.MatrixSpace(self.base_ring(), len(B), self.codomain().rank())
        H = sub.Hom(self.codomain(), self.category_for())
        return H(MS(ims))


class DegeneracyMap(Morphism):
    def __init__(self, parent, A, t, side='left'):
        """
        Create the degeneracy map of index t in parent defined by the
        matrix A.

        INPUT:

        - ``parent`` -- a space of homomorphisms of abelian varieties

        - ``A`` -- a matrix defining self

        - ``t`` -- list of indices defining the degeneracy map

        EXAMPLES::

            sage: J0(44).degeneracy_map(11,2)
            Degeneracy map from Abelian variety J0(44) of dimension 4 to Abelian variety J0(11) of dimension 1 defined by [2]
            sage: J0(44)[0].degeneracy_map(88,2)
            Degeneracy map from Simple abelian subvariety 11a(1,44) of dimension 1 of J0(44) to Abelian variety J0(88) of dimension 9 defined by [2]
        """
        if not isinstance(t, list):
            t = [t]
        self._t = t
        Morphism.__init__(self, parent, A, side)

    def t(self):
        """
        Return the list of indices defining ``self``.

        EXAMPLES::

            sage: J0(22).degeneracy_map(44).t()
            [1]
            sage: J = J0(22) * J0(11)
            sage: J.degeneracy_map([44,44], [2,1])
            Degeneracy map from Abelian variety J0(22) x J0(11) of dimension 3 to Abelian variety J0(44) x J0(44) of dimension 8 defined by [2, 1]
            sage: J.degeneracy_map([44,44], [2,1]).t()
            [2, 1]
        """
        return self._t

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: J0(22).degeneracy_map(44)._repr_()
            'Degeneracy map from Abelian variety J0(22) of dimension 2 to Abelian variety J0(44) of dimension 4 defined by [1]'
        """
        return "Degeneracy map from %s to %s defined by %s" % (self.domain(), self.codomain(), self._t)


class HeckeOperator(Morphism):
    """
    A Hecke operator acting on a modular abelian variety.
    """
    def __init__(self, abvar, n, side='left'):
        """
        Create the Hecke operator of index `n` acting on the
        abelian variety abvar.

        INPUT:

        - ``abvar`` -- a modular abelian variety

        - ``n`` -- positive integer

        EXAMPLES::

            sage: J = J0(37)
            sage: T2 = J.hecke_operator(2); T2
            Hecke operator T_2 on Abelian variety J0(37) of dimension 2
            sage: T2.parent()
            Endomorphism ring of Abelian variety J0(37) of dimension 2
        """
        from .abvar import ModularAbelianVariety_abstract
        n = ZZ(n)
        if n <= 0:
            raise ValueError("n must be positive")
        if not isinstance(abvar, ModularAbelianVariety_abstract):
            raise TypeError("abvar must be a modular abelian variety")
        self.__abvar = abvar
        self.__n = n
        sage.modules.matrix_morphism.MatrixMorphism_abstract.__init__(self, abvar.Hom(abvar), side)

    def _repr_(self):
        """
        String representation of this Hecke operator.

        EXAMPLES::

            sage: J = J0(37)
            sage: J.hecke_operator(2)._repr_()
            'Hecke operator T_2 on Abelian variety J0(37) of dimension 2'
        """
        return "Hecke operator T_%s on %s" % (self.__n, self.__abvar)

    def index(self):
        """
        Return the index of this Hecke operator. (For example, if this is
        the operator `T_n`, then the index is the integer `n`.)

        OUTPUT:

        - ``n`` -- a (Sage) Integer

        EXAMPLES::

            sage: J = J0(15)
            sage: t = J.hecke_operator(53)
            sage: t
            Hecke operator T_53 on Abelian variety J0(15) of dimension 1
            sage: t.index()
            53
            sage: t = J.hecke_operator(54)
            sage: t
            Hecke operator T_54 on Abelian variety J0(15) of dimension 1
            sage: t.index()
            54

        ::

            sage: J = J1(12345)
            sage: t = J.hecke_operator(997) ; t
            Hecke operator T_997 on Abelian variety J1(12345) of dimension 5405473
            sage: t.index()
            997
            sage: type(t.index())
            <class 'sage.rings.integer.Integer'>
        """
        return self.__n

    def n(self):
        r"""
        Alias for ``self.index()``.

        EXAMPLES::

            sage: J = J0(17)
            sage: J.hecke_operator(5).n()
            5
        """
        return self.index()

    def characteristic_polynomial(self, var='x'):
        """
        Return the characteristic polynomial of this Hecke operator in the
        given variable.

        INPUT:

        - ``var`` -- string (default: ``'x'``)

        OUTPUT: a polynomial in var over the rational numbers

        EXAMPLES::

            sage: A = J0(43)[1]; A
            Simple abelian subvariety 43b(1,43) of dimension 2 of J0(43)
            sage: t2 = A.hecke_operator(2); t2
            Hecke operator T_2 on Simple abelian subvariety 43b(1,43) of dimension 2 of J0(43)
            sage: f = t2.characteristic_polynomial(); f
            x^2 - 2
            sage: f.parent()
            Univariate Polynomial Ring in x over Integer Ring
            sage: f.factor()
            x^2 - 2
            sage: t2.characteristic_polynomial('y')
            y^2 - 2
        """
        return self.__abvar.rational_homology().hecke_polynomial(self.__n, var).change_ring(ZZ)

    def charpoly(self, var='x'):
        r"""
        Synonym for ``self.characteristic_polynomial(var)``.

        INPUT:

        - ``var`` -- string (default: ``'x'``)

        EXAMPLES::

            sage: A = J1(13)
            sage: t2 = A.hecke_operator(2); t2
            Hecke operator T_2 on Abelian variety J1(13) of dimension 2
            sage: f = t2.charpoly(); f
            x^2 + 3*x + 3
            sage: f.factor()
            x^2 + 3*x + 3
            sage: t2.charpoly('y')
            y^2 + 3*y + 3
        """
        return self.characteristic_polynomial(var)

    def fcp(self, var='x'):
        """
        Return the factorization of the characteristic polynomial.

        EXAMPLES::

            sage: t2 = J0(33).hecke_operator(2)
            sage: t2.charpoly()
            x^3 + 3*x^2 - 4
            sage: t2.fcp()
            (x - 1) * (x + 2)^2
        """
        return self.charpoly(var).factor()

    def action_on_homology(self, R=ZZ):
        r"""
        Return the action of this Hecke operator on the homology
        `H_1(A; R)` of this abelian variety with coefficients in
        `R`.

        EXAMPLES::

            sage: A = J0(43)
            sage: t2 = A.hecke_operator(2); t2
            Hecke operator T_2 on Abelian variety J0(43) of dimension 3
            sage: h2 = t2.action_on_homology(); h2
            Hecke operator T_2 on Integral Homology of Abelian variety J0(43) of dimension 3
            sage: h2.matrix()
            [-2  1  0  0  0  0]
            [-1  1  1  0 -1  0]
            [-1  0 -1  2 -1  1]
            [-1  0  1  1 -1  1]
            [ 0 -2  0  2 -2  1]
            [ 0 -1  0  1  0 -1]
            sage: h2 = t2.action_on_homology(GF(2)); h2
            Hecke operator T_2 on Homology with coefficients in Finite Field of size 2 of Abelian variety J0(43) of dimension 3
            sage: h2.matrix()
            [0 1 0 0 0 0]
            [1 1 1 0 1 0]
            [1 0 1 0 1 1]
            [1 0 1 1 1 1]
            [0 0 0 0 0 1]
            [0 1 0 1 0 1]
        """
        return self.__abvar.homology(R).hecke_operator(self.index())

    def matrix(self):
        r"""
        Return the matrix of ``self`` acting on the homology
        `H_1(A, ZZ)` of this abelian variety with coefficients in `\ZZ`.

        EXAMPLES::

            sage: J0(47).hecke_operator(3).matrix()
            [ 0  0  1 -2  1  0 -1  0]
            [ 0  0  1  0 -1  0  0  0]
            [-1  2  0  0  2 -2  1 -1]
            [-2  1  1 -1  3 -1 -1  0]
            [-1 -1  1  0  1  0 -1  1]
            [-1  0  0 -1  2  0 -1  0]
            [-1 -1  2 -2  2  0 -1  0]
            [ 0 -1  0  0  1  0 -1  1]

        ::

            sage: J0(11).hecke_operator(7).matrix()
            [-2  0]
            [ 0 -2]
            sage: (J0(11) * J0(33)).hecke_operator(7).matrix()
            [-2  0  0  0  0  0  0  0]
            [ 0 -2  0  0  0  0  0  0]
            [ 0  0  0  0  2 -2  2 -2]
            [ 0  0  0 -2  2  0  2 -2]
            [ 0  0  0  0  2  0  4 -4]
            [ 0  0 -4  0  2  2  2 -2]
            [ 0  0 -2  0  2  2  0 -2]
            [ 0  0 -2  0  0  2  0 -2]

        ::

            sage: J0(23).hecke_operator(2).matrix()
            [ 0  1 -1  0]
            [ 0  1 -1  1]
            [-1  2 -2  1]
            [-1  1  0 -1]
        """
        try:
            return self._matrix
        except AttributeError:
            pass
        self._matrix = self.action_on_homology().matrix()
        return self._matrix
